'use strict';

var lastTap = 0;
var timeout;

// A canvas to pre-render pngs, svgs...
class OffScreenImage {
  constructor(url, sf) {
    this.render = this.render.bind(this);
    this.renderInRect = this.renderInRect.bind(this);
    if(sf == undefined)
      sf = 1;
    this.canvas = document.createElement('canvas');
    var img = new Image;
    img.src = url;
    var that = this;
    img.onload = function() {
      that.nominalWidth = img.width;
      that.nominalHeight = img.height;
      that.width = img.width * sf;
      that.height =  img.height * sf;
      that.canvas.width = 2 * that.width;
      that.canvas.height = 2 * that.height;
      var ctx = that.canvas.getContext("2d");
      ctx.drawImage(img, 0, 0, img.width * 2 * sf, img.height * 2 * sf);
    }
    img.src = url;
  }
  render(c, l, t) {
    c.drawImage(this.canvas, l, t, this.width, this.height);
  }
  renderInRect(c, l, t, w, h) {
    c.save();

    var ws = w / this.width;
    var hs = h / this.height;

    var sf = Math.max(ws, hs);

    c.translate(l, t);
    c.scale(sf, sf);

    c.drawImage(this.canvas, 0, 0, this.width, this.height);
    c.restore();
  }
}

class ScaledSVG {
  constructor(url, targetSize) {
    this.render = this.render.bind(this);
    this.renderInRect = this.renderInRect.bind(this);
    this.canvas = document.createElement('canvas');
    var img = new Image;
    img.src = url;
    img.onload = ()=> {
      const aspectRatio = img.width / img.height;
      if (targetSize == undefined) {
        const w = img.width;
        const h = img.height;

        this.nominalWidth = img.width;
        this.nominalHeight = img.height;
        this.width = w;
        this.height = h;

        this.canvas.width = 2 * w;
        this.canvas.height = 2 * h;

        let ctx = this.canvas.getContext("2d");
        ctx.drawImage(img, 0, 0, w *2, h * 2);
      }
      else if (targetSize.height != undefined) {
        const h = targetSize.height;
        const w = h * aspectRatio;

        this.nominalWidth = img.width;
        this.nominalHeight = img.height;
        this.width = w;
        this.height = h;

        this.canvas.width = 2 * w;
        this.canvas.height = 2 * h;

        let ctx = this.canvas.getContext("2d");
        ctx.drawImage(img, 0, 0, w *2, h * 2);
      }
      else if (targetSize.width != undefined) {
        const w = targetSize.width;
        const h = w * (1.0 / aspectRatio);

        this.nominalWidth = img.width;
        this.nominalHeight = img.height;
        this.width = w;
        this.height = h;

        this.canvas.width = 2 * w;
        this.canvas.height = 2 * h;
        let ctx = this.canvas.getContext("2d");
        ctx.drawImage(img, 0, 0, w *2, h * 2);
      }
    }
    img.src = url;
  }

  render(c, l, t) {
    c.drawImage(this.canvas, l, t, this.width, this.height);
  }
  renderInRect(c, l, t, w, h) {
    c.save();

    var ws = w / this.width;
    var hs = h / this.height;

    var sf = Math.max(ws, hs);

    c.translate(l, t);
    c.scale(sf, sf);

    c.drawImage(this.canvas, 0, 0, this.width, this.height);
    c.restore();
  }
}

// Root view that installs event handlers and has a real canvas
class CanvasRoot {
  constructor(animator, noAttach) {

    this.animator = animator;
    // Events
    this.dblClick = this.dblClick.bind(this);
    this.dblTap = this.dblTap.bind(this);
    this.mouseBegin = this.mouseBegin.bind(this);
    this.mousePerform = this.mousePerform.bind(this);
    this.mouseEnd = this.mouseEnd.bind(this);
    this.resized = this.resized.bind(this);
    // Drawing
    this.render = this.render.bind(this);
    this.derivedRender = this.derivedRender.bind(this);
    this.requestRender = this.requestRender.bind(this);
    this.requestRenderAll = this.requestRenderAll.bind(this);
    this.getMouseHandler = this.getMouseHandler.bind(this);
    // Root
    this.scale = this.scale.bind(this);
    this.setVisibleRect = this.setVisibleRect.bind(this);
    this.getRootDiv = this.getRootDiv.bind(this);
    this.addSub = this.addSub.bind(this);
    this.removeSub = this.removeSub.bind(this);
    this.setVis = this.setVis.bind(this);

    this.r = new Rect(0, 0, 0, 0); // this is the visible rect with the canvas, not the container rect
    this.rootDiv = null;
    this.subViews = new List;
    this.renderList = new List;
    this.divScale = 1.0;
    this.bg = kMixBG;

    this.canvas = document.createElement('canvas');
    this.canvas.style.position = "fixed";
    this.canvas.style.pointerEvents = "none";
    this.context = this.canvas.getContext("2d", { alpha: false });
    this.updateCanvasSize = false;

    this.div = document.createElement('div');
    this.div.id = "cvRoot";
    this.div.addEventListener('dblclick', this.dblClick);
    this.div.addEventListener('touchend', this.dblTap);
    this.div.addEventListener('mousedown', this.mouseBegin);
    this.div.addEventListener('touchstart', this.mouseBegin);
    this.div.addEventListener('contextmenu', e => { e.preventDefault(); });
    this.div.style.transformOrigin = "top";

    // Most of the time we want to attach automatically (except scroller)
    if(noAttach == undefined) {
      this.rootDiv = this.div;
      this.rootDiv.appendChild(this.canvas);
    }
  }

  setBG(bg) {
    this.bg = bg;
  }

  addSub(v) {
    this.subViews.add(v);
    v.setRoot(this);
  }
  removeSub(v) {
    v.setRoot(undefined);
    this.subViews.remove(v);
    this.requestRenderAll();
  }

  scale(v) {
    if(v != this.divScale) {
      this.divScale = v;
      this.div.style.transform = "scale(" + v + ")";
    }
  }
  setVisibleRect(l, t, w, h) {
    this.r.setLTWH(l,t,w,h);
    this.updateCanvasSize = true;
    this.requestRenderAll();
  }
  setVis(isVis) {
    var n = this.subViews.length();
    for(var i = 0; i < n; i++)
      this.subViews.get(i).setParentVis(isVis);
    if(isVis)
      this.requestRenderAll();
  }

  getRootDiv() { return this.rootDiv; }

  requestRenderAll() {
    this.renderList.items = this.subViews.items.slice();
    this.animator.queue(this.render);
  }

  requestRender(v) {
    if(this.r.intersects(v.r)) {
      this.renderList.add(v);
      if(this.renderList.length() == 1)
          this.animator.queue(this.render);
    }
  }

  render() {
    if(this.updateCanvasSize) {
      this.updateCanvasSize = false;
      var sf = kCanvasScale;
      this.canvas.style.width = this.r.w.toFixed() + "px";
      this.canvas.style.height = this.r.h.toFixed() + "px";
      this.canvas.width = sf * this.r.w;
      this.canvas.height = sf * this.r.h;
      this.context.scale(sf, sf);
      //this.context.translate(0.5, 0.5);
      this.context.fillStyle = this.bg;
      this.context.fillRect(0, 0, this.r.w, this.r.h)
    }

    this.context.save();

    // draw all the views in the renderList
    this.context.translate(- this.r.l, - this.r.t);

    this.context.save();
    this.derivedRender(this.context);
    this.context.restore();

    var n = this.renderList.length();
    for(var i = 0; i < n; i++) {
      var v = this.renderList.get(i);
      if(this.r.intersects(v.r)) {
        const cumulativeAlpha = v.parentView ? v.parentView.getCumulativeEffectiveAlpha() : 1.0;

        this.context.save();

        if (cumulativeAlpha != 1.0) {

          // this.context.clip(v.l, v.t, v.w, v.h);
          const oldFillStyle = this.context.fillStyle;
          this.context.fillStyle = this.bg;
          this.context.fillRect(v.l, v.t, v.w, v.h);
          this.context.globalAlpha = cumulativeAlpha;
          this.context.fillStyle = oldFillStyle;
        }
        v.renderSelfAndSubs(this.context);
        this.context.restore();
      }
    }
    this.context.restore();

    this.renderList.clear();
  }

  derivedRender(c) {}

  resized() {
    var cs = getComputedStyle(this.div);
    var w = parseInt(cs.width, 10);
    var h = parseInt(cs.height, 10);
    if(isNaN(w))
      w = 0;
    if(isNaN(h))
      h = 0;

    this.setVisibleRect(this.div.scrollLeft, this.div.scrollTop, w, h);
  }
  getMouseHandler(xy) {
    if(this.r.contains(xy.x, xy.y)) {
      // find the topmost view that contains the coordinate
      var n = this.subViews.length();
      for(var i = n-1; i >= 0; i--) {
        var v = this.subViews.get(i).getMouseHandler(xy);
        if(v != undefined)
          return v;
      }
    }
    return undefined;
  }
  dblTap(e) {
    const curTime = new Date().getTime();
    const tapLen = curTime - lastTap;
    if (tapLen < 300 && tapLen > 0) {
      this.dblClick(e)
    } else {
      timeout = setTimeout(() => {
        clearTimeout(timeout);
      }, 300);
    }
    lastTap = curTime;
  }
  dblClick(e) {
    this.mouseEnd(e);
    this.startXY = this.mouseGetXY(e);

    var mh = this.getMouseHandler(this.startXY);
    if(mh != undefined) {
      e.preventDefault();
      mh.handleDblClick();
    }
  }
  mouseBegin(e) {
    if((e.button == 0) || e.touches) {
      this.mouseEnd(e);
      this.startXY = this.mouseGetXY(e);

      this.mouseHandler = this.getMouseHandler(this.startXY);
      if(this.mouseHandler != undefined) {
        document.addEventListener('mousemove', this.mousePerform);
        document.addEventListener('touchmove', this.mousePerform);
        document.addEventListener('mouseup', this.mouseEnd);
        document.addEventListener('touchend', this.mouseEnd);

        e.preventDefault();
        this.mouseHandler.handleMouseDown(this.startXY, e);
      }
    }
  }
  mousePerform(e) {
    if(this.mouseHandler != undefined) {
      var xy = this.mouseGetXY(e);
      this.mouseHandler.handleMouseMoved(xy, this.mouseDeltaXY(this.startXY, xy), e);
    }
  }
  mouseEnd(e) {
    if(this.mouseHandler != undefined) {
      document.removeEventListener('mousemove', this.mousePerform);
      document.removeEventListener('touchmove', this.mousePerform);
      document.removeEventListener('mouseup', this.mouseEnd);
      document.removeEventListener('touchend', this.mouseEnd);

      this.mouseHandler.handleMouseUp(this.mouseGetXY(e));
    }
    this.mouseHandler = undefined;
  }
  mouseGetXY(e) {
    // Apparently getting the mouse relatove to the canvas is not trivial
    // See https://stackoverflow.com/questions/17130395/real-mouse-position-in-canvas
    var rd = this.getRootDiv();
    var r = rd.getBoundingClientRect();
    return {
      x: ((e.changedTouches ? e.changedTouches[0].clientX : e.x) - (r.left - rd.scrollLeft)) / this.divScale,
      y: ((e.changedTouches ? e.changedTouches[0].clientY : e.y) - (r.top - rd.scrollTop)) / this.divScale
    };
  }
  mouseDeltaXY(start, end) {
    var x = end.x - start.x;
    var y = start.y - end.y;
    return {x, y};
  }
}


class ModalCanvas extends CanvasRoot {
  constructor(appdata, autoDeactivate) {
    super(appdata.animator);

    this.activate = this.activate.bind(this);
    this.deactivate = this.deactivate.bind(this);
    this.setBG = this.setBG.bind(this);
    this.filterKey = this.filterKey.bind(this);
    this.keyFilterer = appdata.keyFilter;

    this.clearColor = kMixBG;

    // the regular div will be placed in the modal div.
    this.div.style.position = "absolute";
    this.div.style.backgroundColor = kMixBG;
    this.div.style.top = "0px";
    this.div.style.left = "0px";
    this.div.style.width = "0px";
    this.div.style.height = "0px";
    //this.div.style.float = "left";

    this.mdiv = document.createElement('div');
    this.mdiv.id = "modalcontent";
    this.mdiv.style.position = "absolute";
    this.mdiv.style.top = "0px";
    this.mdiv.style.left = "0px";
    this.mdiv.style.width = "100%";
    this.mdiv.style.height = "100%";
    this.mdiv.style.backgroundColor = "rgba(0,0,0,0.6)";
    this.mdiv.style.float = "left";
    this.mdiv.style.zIndex = 1000;
    this.mdiv.appendChild(this.div)

    if(autoDeactivate) {
      this.mouseDeactivate = this.mouseDeactivate.bind(this);
      this.mdiv.addEventListener('mousedown', this.mouseDeactivate);
      this.mdiv.addEventListener('touchstart', this.mouseDeactivate);
    }

    // this will be the rect of the modal div (the window)
    this.wr = new Rect(0,0,0,0);
    this.mv = null;

    appdata.eventSys.addListener('wResized', this.resized);
  }
  setBG(color) {
    this.div.style.backgroundColor = color;
    this.clearColor = color;
  }
  activate(v) {
    this.deactivate();
    this.keyFilterer.add(this);
    document.body.appendChild(this.mdiv);
    this.mv = v;
    this.addSub(v);
    this.r.setLTWH(0, 0, v.w, v.h);
    this.updateCanvasSize = true;
    this.requestRenderAll();
    this.resized();
  }
  deactivate() {
    this.renderList.clear();
    if(this.mv != null) {
      document.body.removeChild(this.mdiv);
      this.removeSub(this.mv);
      this.mv = null;
      this.keyFilterer.remove();
    }
  }
  mouseDeactivate(e) {
    // look for mouse clicks that are not in the content div and deactivate
    var xy = this.mouseGetXY(e);
    if(this.r.contains(xy.x, xy.y))
      return;

    this.deactivate();
  }
  getMouseHandler(xy) {
    if(this.r.contains(xy.x, xy.y))
      return super.getMouseHandler(xy);

    return null;
  }
  filterKey(keycode) {}
  resized() {

    var cs = getComputedStyle(this.mdiv);
    var w = parseInt(cs.width, 10);
    var h = parseInt(cs.height, 10);
    if(isNaN(w))
      w = 0;
    if(isNaN(h))
      h = 0;

    this.setVisibleRect(0, 0, w, h);

    // keep the modal window scaled inside the container with a border of 20
    var sfw = (w- 20)/this.r.w;
    var sfh = (h - 20)/this.r.h;
    // todo, on ios allow the modal canvas to scale past 1

    var sf = Math.min(sfw, sfh);
    // only scale down on desktop browsers
    // if(!isMobile)
      sf = Math.min(1, sf);

    this.scale(sf);
  }
  setVisibleRect(l, t, w, h) {
    if(this.mv != null) {
      this.wr.setLTWH(l,t,w,h);
      this.updateCanvasSize = true;
      this.requestRenderAll();
    }
  }

  render() {
    if(this.mv != null) {

      if(this.updateCanvasSize) {
        this.updateCanvasSize = false;
        var sf = kCanvasScale;
        var canvasL = this.wr.w/2 - this.r.w/2;
        this.div.style.top = "10px";
        this.div.style.left = canvasL + "px";
        this.div.style.width = this.canvas.style.width = this.r.w.toFixed() + "px";
        this.div.style.height = this.canvas.style.height = this.r.h.toFixed() + "px";
        this.canvas.width = sf * this.r.w;
        this.canvas.height = sf * this.r.h;
        this.context.scale(sf, sf);
        //this.context.translate(0.5, 0.5);
        this.context.fillStyle = this.clearColor;
        this.context.fillRect(0, 0, this.r.w, this.r.h)
      }

      // draw all the views in the renderList
      //this.context.translate(- this.r.l, - this.r.t);
      var n = this.renderList.length();
      for(var i = 0; i < n; i++) {
        var v = this.renderList.get(i);
        if(this.r.intersects(v.r))
          v.renderSelfAndSubs(this.context);
      }

      this.renderList.clear();
    }
  }
}

