

class DeviceListItemView {
  constructor(dev) {
    this.attachTo = this.attachTo.bind(this);
    this.detach = this.detach.bind(this);
    this.clickHandler = this.clickHandler.bind(this);
    this.isActive = this.isActive.bind(this);
    this.updateClass = this.updateClass.bind(this);

    this.device = dev;

    this.div = document.createElement("div");

    this.devNameP = this.makeParagraph(dev.getFriendlyName(), "0px", "0px", "22px Arial");
    this.snP = this.makeParagraph("SN: " + dev.serial, "0px", "0px", "12px Arial");

    const hostString = dev.isLocalDevice() ? "Local" : dev.hostname;
    this.hostP = this.makeParagraph("Host: " + hostString, "0px", "30px", "12px Arial");

    for (const p of [this.devNameP, this.snP, this.hostP]) {
      this.div.appendChild(p);
    }

    this.div.addEventListener("click", this.clickHandler);
    this.updateClass();
  }

  makeParagraph(text, topPadding, bottomPadding, font, color) {
    let p = document.createElement("p");
    let t = document.createTextNode(text);

    p.style.textAlign = "center";
    p.style.paddingLeft = "10px";
    p.style.paddingRight = "10px";
    p.style.marginTop = "0px";
    p.style.marginBottom = "0px";
    p.style.paddingTop = topPadding;
    p.style.paddingBottom = bottomPadding;
    p.style.font = font;
    p.style.userSelect = "none";
    p.style.webkitUserSelect = "none";
    p.style.cursor = "default";

    if (color != undefined)
      p.style.color = color;

    p.appendChild(t);
    return p;
  }

  detach() {
    if (this.parentDiv != undefined) {
      this.parentDiv.removeChild(this.div);
    }
    gApp.appdata.eventSys.removeListener('ProxyStatusChanged', this.updateClass);
    gApp.appdata.eventSys.removeListener('DeviceStatusChanged', this.updateClass);
    gApp.appdata.eventSys.removeListener("DeviceAuthenticated", this.updateClass);
  }
  attachTo(div) {
    this.detach();
    this.parentDiv = div;
    this.parentDiv.appendChild(this.div);

    // If the device disconnects, change the text color.
    gApp.appdata.eventSys.addListener('ProxyStatusChanged', this.updateClass);
    gApp.appdata.eventSys.addListener('DeviceStatusChanged', this.updateClass);
    gApp.appdata.eventSys.addListener("DeviceAuthenticated", this.updateClass);
  }

  clickHandler(e) {
    const device = gApp.getActiveDevice();
    if (device != undefined && device.isEqualTo(this.device)) {
      HideDiscoveryView();
      return;
    }
    gApp.setActiveDevice(this.device);
  }

  isActive() {
    const dev = gApp.appdata.device.getConnectedDevice();
    return dev != undefined && dev.isEqualTo(this.device) && gApp.appdata.device.authenticated == true;
  }
  updateClass() {
    this.div.className = this.isActive() ?  "discovery-device-active" : "discovery-device-normal";
  }
};


class ContextMenu {
  constructor(d, opts) {
    this.syncPosition = this.syncPosition.bind(this);
    this.makeOptionDiv = this.makeOptionDiv.bind(this);
    this.show = this.show.bind(this);
    this.hide = this.hide.bind(this);
    this.isShowing = this.isShowing.bind(this);

    this.masterDiv = d;

    this.div = document.createElement("div");
    this.div.id = "ContextMenu";
    this.div.style.position = "absolute";

    this.div.style.width = "fit-content";
    this.div.style.height = "fit-content";

    this.div.style.display = "none";
    this.div.style.overflow = "hidden";
    this.div.style.overflowY = "auto";
    this.div.style.backgroundColor = "rgb(255, 255, 255)";
    this.div.style.zIndex = kZIndexContextMenu;
    this.div.style.boxShadow = "0 4px 8px 0 rgba(0, 0, 0, 0.2), 0 6px 20px 0 rgba(0, 0, 0, 0.19)";

    for (const o of opts) {
      this.div.appendChild(this.makeOptionDiv(o));
    }

    gApp.appdata.eventSys.addListener("wResized", this.syncPosition);
    this.observer = new MutationObserver((mutationsList, observer) => {
      this.syncPosition();
    });

    this.observer.observe(document.body, this.mutationObserverConfig());

    this.masterDiv.addEventListener("click", (e)=>{
      e.stopImmediatePropagation();
      if (this.isShowing())
        this.hide();
      else
        this.show();
    });
  }

  mutationObserverConfig() {
    return { attributes: true, childList: true, subtree: true, attributeFilter: ["style"] };
  }

  makeOptionDiv(o) {
    var odiv = document.createElement("div");
    odiv.className = "context-menu-text";
    odiv.style.width = "fit-content";
    odiv.style.height = "fit-content";
    odiv.style.textAlign = "center";
    odiv.style.minWidth = "100%";
    odiv.style.display = "flex";

    var tdiv = document.createElement("div");
    var t = document.createTextNode(o.text);
    tdiv.appendChild(t);

    odiv.appendChild(tdiv);

    odiv.addEventListener("click", o.callback);
    return odiv;
  }
  syncPosition(){
    this.observer.disconnect(); // Or else we'll get called recursively

    // This is a massive kludge, but I can't figure out how to achieve this more elegantly.
    // If the bottom of the context menu extends past the bottom of the viewport,
    // I want to shrink the height of the menu so the overflow enables scrolling.
    // But I also want to "fit-content" if the viewport height becomes large enough
    // to accomodate the natural height of the menu. I will try to store the "natural size" of
    // the menu (i.e. the size computed with "fit-content") in this.computedStyle.
    // But I'm not sure exactly when that style can be computed, so I'm assuming the maximum value we ever
    // see is the "natural size".
    let cs = getComputedStyle(this.div);

    if (cs == undefined)
      return;

    let currentWidth = Number(cs.width.replace("px", ""));
    let currentHeight = Number(cs.height.replace("px", ""));

    if (isNaN(currentWidth) || isNaN(currentHeight))
      return; // Maybe we haven't been added to the document yet.

    if (this.computedStyle == undefined) {
      this.computedStyle = {
        width: currentWidth,
        height: currentHeight
      };
    }
    else {
      if (currentWidth > this.computedStyle.width)
        this.computedStyle.width = currentWidth;
      if (currentHeight > this.computedStyle.height)
        this.computedStyle.height = currentHeight;
    }

    // Try to place this menu underneath wherever the "master div" is
    // within the viewport, but also constrain it so that it's on screen,
    // even if the "master div" is not.
    let targetRect = this.masterDiv.getBoundingClientRect();
    let cx = targetRect.left + (targetRect.width / 2);
    let myRect = this.div.getBoundingClientRect();
    let x = cx - (myRect.width / 2);
    let y = targetRect.bottom;

    let r = x + this.computedStyle.width;
    const kPad = 20;
    if (r > document.body.clientWidth + kPad) {
      x -= (r - document.body.clientWidth + kPad);
    }

    if (x < kPad)
      x = kPad;

    let dx = x - myRect.x;
    let dy = y - myRect.y;

    let x0 = Number(this.div.style.left.replace("px", ""));
    let y0 = Number(this.div.style.top.replace("px", ""));

    this.div.style.left = String(x0 + dx) + "px";
    this.div.style.top = String(y0 + dy) + "px";

    let b = y + this.computedStyle.height;

    if (b > document.body.clientHeight) {
      this.div.style.maxHeight = String(document.body.clientHeight - y) + "px";
    } 
    else {
      this.div.style.maxHeight = "fit-content";
    }

    // Reconnect the observer now that we've made changes.
    this.observer.observe(document.body, this.mutationObserverConfig());
  }

  show() {
    if (this.isShowing())
      return;

    // clicking on the cancel div will dismiss the menu
    this.cancelDiv = document.createElement("div");
    this.cancelDiv.id = "ContextMenuCancelDiv";
    this.cancelDiv.style.position = "absolute";
    this.cancelDiv.style.pointerEvents = "auto";
    this.cancelDiv.style.width = "100%";
    this.cancelDiv.style.height = "100%";
    this.cancelDiv.style.backgroundColor = "rgba(0,0,0,0)";
    this.cancelDiv.style.zIndex = kZIndexConsteMenuCanceler;

    this.cancelDiv.addEventListener("click", (e)=>{
      e.stopImmediatePropagation();
      this.hide();
    });

    document.addEventListener("keyup", (e)=>{
      if (e.key == "Escape") {
        e.stopImmediatePropagation();
        this.hide();
      }
    });

    document.body.appendChild(this.cancelDiv);

    // We want a vertical stack of options.
    this.div.style.display = "flex";
    this.div.style.flexDirection = "column";
    this.syncPosition();
  }

  hide() {
    if (this.cancelDiv != undefined) {
      document.body.removeChild(this.cancelDiv);
      this.cancelDiv = undefined;
    }
    this.div.style.display = "none";
  }

  close() {
    if (document.body.contains(this.div))
      document.body.removeChild(this.div);
    if (document.body.contains(this.cancelDiv))
      document.body.removeChild(this.cancelDiv);
    this.cancelDiv = undefined;
  }

  isShowing() {
    return this.div.style.display != "none";
  }

};
const kBackButtonSize = 25;

class BackButton {
  constructor() {
    this.updateVisibility = this.updateVisibility.bind(this);

    var innerDiv = document.createElement("div");
    innerDiv.style.display = "flex";
    innerDiv.style.alignItems = "center"; // Center vertically
    innerDiv.style.justifyContent = "center"; // Center horizontally
    innerDiv.style.borderColor = kTabTxt;
    innerDiv.style.color = kTabTxt;
    innerDiv.style.backgroundColor = "rgba(0,0,0,0)";
    innerDiv.style.width = String(kBackButtonSize) + "px";
    innerDiv.style.height = String(kBackButtonSize) + "px";
    innerDiv.style.textAlign = "center";
    innerDiv.style.boxSizing = "border-box"
    innerDiv.style.pointerEvents = "auto";

    this.innerDiv = innerDiv;

    this.div = document.createElement("div");

    this.div.id = "DeviceListBackButton";
    this.div.style.display = "flex";
    this.div.style.alignItems = "center"; // Center vertically
    this.div.style.justifyContent = "center"; // Center horizontally
    this.div.style.borderColor = kTabTxt;
    this.div.style.color = kTabTxt;
    this.div.style.backgroundColor = "rgba(0,0,0,0)";
    this.div.style.width = String(kBackButtonSize) + "px";
    this.div.style.height = "100%";
    this.div.style.textAlign = "center";
    this.div.style.paddingLeft = "50px";
    this.div.style.pointerEvents = "none";

    this.image = document.createElement("img");
    this.image.src = "back.svg";
    this.image.style.maxWidth = "100%";
    this.image.style.maxHeight = "100%";
    this.image.style.minWidth = "100%";
    this.image.style.minHeight = "100%";
    this.image.style.pointerEvents = "none";

    innerDiv.appendChild(this.image);

    this.div.appendChild(innerDiv);

    gApp.appdata.eventSys.addListener("AppVisibilityChanged", this.updateVisibility);
    this.updateVisibility();
  }

  updateVisibility() {
    this.div.style.visibility = gApp.visible ? "visible" : "hidden";
  }
};

class SettingsDiv {
  constructor() {

    this.ipConnect = this.ipConnect.bind(this);
    this.connectVirtual828 = this.connectVirtual828.bind(this);
    this.connectVirtualUL5 = this.connectVirtualUL5.bind(this);
    this.logoutAll = this.logoutAll.bind(this);
    this.enableWebMIDI = this.enableWebMIDI.bind(this);

    var innerDiv = document.createElement("div");
    innerDiv.style.display = "flex";
    innerDiv.style.alignItems = "center"; // Center vertically
    innerDiv.style.justifyContent = "center"; // Center horizontally
    innerDiv.style.backgroundColor = "rgba(0,0,0,0)";
    innerDiv.style.width = String(kBackButtonSize) + "px";
    innerDiv.style.height = String(kBackButtonSize) + "px";
    innerDiv.style.boxSizing = "border-box"
    innerDiv.style.pointerEvents = "auto";

    this.innerDiv = innerDiv;

    this.div = document.createElement("div");

    this.div.id = "DiscoverySettings";
    this.div.style.display = "flex";
    this.div.style.alignItems = "center"; // Center vertically
    this.div.style.justifyContent = "center"; // Center horizontally
    this.div.style.backgroundColor = "rgba(0,0,0,0)";
    this.div.style.width = String(kBackButtonSize) + "px";
    this.div.style.height = "100%";
    this.div.style.textAlign = "center";
    this.div.style.paddingRight = "50px";
    this.div.style.pointerEvents = "none";

    this.image = document.createElement("img");
    this.image.src = "settings_.svg";
    this.image.style.maxWidth = "100%";
    this.image.style.maxHeight = "100%";
    this.image.style.minWidth = "100%";
    this.image.style.minHeight = "100%";
    this.image.style.pointerEvents = "none";

    innerDiv.appendChild(this.image);

    this.div.appendChild(innerDiv);

    let settingsOptions = [
      {
        text: "Connect to IP address...",
        callback: this.ipConnect
      },
      {
        text: "Connect to virtual 828",
        callback: this.connectVirtual828
      },
      {
        text: "Connect to virtual Ultralite-mk5",
        callback: this.connectVirtualUL5
      },
      {
        text: "Log out of all devices",
        callback: this.logoutAll
      }
    ];

    if (PlatformSupportsWebMIDI() && !gApp.appdata.device.jsHostMsgPipe) {
      settingsOptions.splice(settingsOptions.length - 1, 0, {
        text: "Enable WebMIDI",
        callback: this.enableWebMIDI
      });
    }

    this.settingsMenu = new ContextMenu(this.innerDiv, settingsOptions);
    document.body.appendChild(this.settingsMenu.div);

    // Click anywhere will dismiss the menu
    document.body.addEventListener("click", (e)=>{
      if (this.settingsMenu.isShowing() && !this.settingsMenu.div.contains(e.target)) {
        this.settingsMenu.hide();
        e.preventDefault();
        e.stopImmediatePropagation();
      }
    });
  }

  ipConnect() {
    console.log("ipConnect");
    this.settingsMenu.hide();

    let w = 400;
    let h = 250;

    let cx = document.body.clientWidth / 2.0;
    let cy = document.body.clientHeight / 2.0;

    let x = cx - (w/2.0);
    let y = cy - (h/2.0);

    const callbacks = {
      completion: (v, dialog) =>{
        if (dialog != undefined && typeof dialog.dismiss == "function")
          dialog.dismiss();
        else {
          console.log(dialog);
        }
        gApp.appdata.device.disconnect();
        gApp.appdata.device.wsConnect(v);
      },
      canceled: (dialog) => {
        dialog.dismiss();
      }
    };

    new TextEntryDialog(x,y,w,h, gApp.appdata, "Enter IP Address", "", callbacks, false);
  }

  logoutAll() {
    console.log("logoutAll");
    gApp.appdata.device.logout();
    (new PWCache()).clear();
    this.settingsMenu.hide();
  }

  connectVirtual828() {
    console.log("connectVirtual828");
    this.settingsMenu.hide();
    gApp.appdata.networkDeviceList.discoveryAdd(btoa(JSON.stringify(Test828Dict)));
    let dev = gApp.appdata.networkDeviceList.deviceList.find((x)=>{
      return x.ip == Test828Dict.ip;
    });

    if (dev != undefined) {
      gApp.setActiveDevice(dev);
      HideDiscoveryView();
    }
  }
  connectVirtualUL5() {
    console.log("connectVirtualUL5");
    this.settingsMenu.hide();
    gApp.appdata.networkDeviceList.discoveryAdd(btoa(JSON.stringify(TestUL5Dict)));
    let dev = gApp.appdata.networkDeviceList.deviceList.find((x)=>{
      return x.ip == TestUL5Dict.ip;
    });

    if (dev != undefined) {
      gApp.setActiveDevice(dev);
      HideDiscoveryView();
    }
  }
  enableWebMIDI() {
    this.settingsMenu.hide();
    if (PlatformSupportsWebMIDI()) {
      gApp.appdata.device.webMIDI.enableWebMIDI();
    }
  }
};


class DeviceListHeader {
  constructor() {
    this.div = document.createElement("div");
    this.div.id = "DeviceListHeader";

    this.div.style.position = "relative";
    this.div.style.width = "100%";
    this.div.style.height = "100px";
    this.div.style.display = "flex";
    this.div.style.flexDirection = "row";
    this.div.style.color = "rgb(255, 255, 255)";
    this.div.style.textAlign = "center;"
    this.div.style.margin = "auto";
    this.div.style.font = "18pt Arial";
    this.div.style.pointerEvents = "none";

    this.backButton = new BackButton();
    this.div.appendChild(this.backButton.div);

    var textDiv = document.createElement("div");
    textDiv.style.position = "relative";
    textDiv.style.left = "0px";
    textDiv.style.top = "0px";
    textDiv.style.width = "100%";
    textDiv.style.height = "100%";
    textDiv.style.textAlign = "center";
    textDiv.style.margin = "auto";

    var textFlexDiv = document.createElement("div");
    textFlexDiv.style.position = "relative";
    textFlexDiv.style.display = "flex";
    textFlexDiv.style.alignItems = "center"; // Center vertically
    textFlexDiv.style.justifyContent = "center"; // Center horizontally
    textFlexDiv.style.width = "100%";
    textFlexDiv.style.height = "100%";

    var text = document.createTextNode("Available Devices");
    textFlexDiv.appendChild(text);

    textDiv.appendChild(textFlexDiv);

    this.div.appendChild(textDiv);

    this.backButton.innerDiv.addEventListener("click", ()=>{
      HideDiscoveryView();
    });

    this.settingsDiv = new SettingsDiv();
    this.div.appendChild(this.settingsDiv.div);
  }
};

class BrandingDiv {
  constructor() {
    this.div = document.createElement("div");
    this.div.id = "BrandingDiv";

    this.div.style.position = "relative";
    this.div.style.display = "flex";
    this.div.style.alignItems = "center"; // Center vertically
    this.div.style.justifyContent = "center"; // Center horizontally
    this.div.style.width = "100%";
    this.div.style.height = "fit-content";
    this.div.style.margin = "auto";
    this.div.style.font = "18pt Arial";
    this.div.style.pointerEvents = "none";
    this.div.style.paddingTop = "20px";

    this.iconDiv = document.createElement("div");

    this.iconDiv.style.position = "relative";
    this.iconDiv.style.display = "flex";
    this.iconDiv.style.alignItems = "center"; // Center vertically
    this.iconDiv.style.justifyContent = "center"; // Center horizontally

    this.iconDiv.style.width = "128px";
    this.iconDiv.style.height = "128px";

    this.image = document.createElement("img");
    this.image.src = "cm5_256x256.png";
    this.image.style.maxWidth = "100%";
    this.image.style.maxHeight = "100%";

    this.iconDiv.appendChild(this.image);
    this.div.appendChild(this.iconDiv);
  }
};

class DeviceList {
  constructor(l, t, w, h, appdata) {
    this.devListChangeNotification = this.devListChangeNotification.bind(this);
    this.touchesStarted = this.touchesStarted.bind(this);
    this.touchesMoved = this.touchesMoved.bind(this);
    this.touchesEnded = this.touchesEnded.bind(this);

    this.appdata = appdata;
    this.devices = [];
    this.deviceViews = [];

    this.mainDiv = document.createElement("div");
    this.mainDiv.id = "DeviceListContainer";
    this.mainDiv.style.position = "absolute";
    this.mainDiv.style.width = "100%";
    this.mainDiv.style.height = "100%";
    this.mainDiv.style.left = String(l) + "px";
    this.mainDiv.style.top = String(t) + "px";
    this.mainDiv.style.display = "flex";
    this.mainDiv.style.flexDirection = "column";
    this.mainDiv.style.backgroundColor = "rgba(25,25,25,255)";

    this.mainDiv.style.zIndex = kZIndexDiscovery;

    this.branding = new BrandingDiv();
    this.mainDiv.appendChild(this.branding.div);

    this.header = new DeviceListHeader();
    this.mainDiv.appendChild(this.header.div);

    this.div = document.createElement("div");
    this.div.id = "DeviceList";

    this.div.style.overflow = "hidden";
    this.div.style.overflowY = "auto";
    this.div.style.WebkitOverflowScrolling = "touch";
    this.div.style.position = "relative";

    this.div.style.width = "100%";
    this.div.style.height = "100%";

    this.innerDiv = document.createElement("div");
    this.innerDiv.style.width = "100%";
    this.innerDiv.style.height = "fit-content";
    this.div.style.WebkitOverflowScrolling = "touch";

    appdata.eventSys.addListener('NetworkDeviceListChanged', this.devListChangeNotification);

    this.div.appendChild(this.innerDiv);
    this.mainDiv.appendChild(this.div);

    this.mainDiv.addEventListener("touchstart", this.touchesStarted);
    this.mainDiv.addEventListener("touchmove", this.touchesMoved);
    this.mainDiv.addEventListener("touchend", this.touchesEnded);
    this.mainDiv.addEventListener("touchcancel", this.touchesEnded);

    this.devListChangeNotification();

    this.appdata.discoveryView = this;
  }

  touchesStarted(e) {
    if (!gApp.visible)
      return;

    let x = e.changedTouches[0].clientX;
    let y = e.changedTouches[0].clientY;
    let xy = {
      x : x,
      y : y
    };

    if (x >= document.body.clientWidth - kDiscoveryGestureRightEdgeInset && x <= document.body.clientWidth) {
      this.dragTask = new DragTask(xy, ()=>{
        let delta = this.dragTask.getDelta();
        let initial = 0;
        let pos = Math.min(initial, initial + delta.x);
        this.mainDiv.style.left = String(pos) + "px";
      });

      this.swipeDetector = new SwipeDetector(xy);
    }
  }

  touchesMoved(e) {
    if (!gApp.visible)
      return;

    let x = e.changedTouches[0].clientX;
    let y = e.changedTouches[0].clientY;
    let xy = {
      x : x,
      y : y
    };

    if (this.dragTask != undefined) {
      this.dragTask.updateDrag(xy);
    }
    if (this.swipeDetector != undefined) {
      this.swipeDetector.updateDrag(xy);
    }
  }

  touchesEnded(e) {
    if (!gApp.visible)
      return;

    let x = e.changedTouches[0].clientX;
    let y = e.changedTouches[0].clientY;
    let xy = {
      x : x,
      y : y
    };

    if (this.swipeDetector != undefined) {
      this.swipeDetector.touchEnded(xy);

      let swipe = this.swipeDetector.getSwipe();
      if (swipe == Swipe.Left) {
        this.animateSelf(false);
        this.swipeDetector = undefined;
        this.dragTask = undefined;
        return;
      }
      else {
        this.swipeDetector = undefined;
      }
    }

    if (this.dragTask != undefined) {
      let delta = this.dragTask.getDelta();
      if (delta.x < -document.body.clientWidth/2) {
        // The user dismissed the discovery view.
        this.animateSelf(false);
      }
      else {
        this.animateSelf(true);
      }
    }

    this.swipeDetector = undefined;
    this.dragTask = undefined;
  }

  animateSelf(animateIn) {
    var targetLeft = 0;
    if (!animateIn) {
      targetLeft = -document.body.clientWidth;
    }

    const canAnimate = this.mainDiv.animate && typeof this.mainDiv.animate == "function";
    if (!canAnimate) {
      if (animateIn) {
        this.mainDiv.style.left = String(targetLeft) + "px";
      }
      else {
        document.body.removeChild(this.mainDiv);
      }
    }
    else {
      let animation = this.mainDiv.animate(
        [
          {
            // from
          },
          {
            // to
            left: String(targetLeft) + "px"
          }
        ],
        {
          duration: kDiscoverySwipeDuration,
          easing: 'cubic-bezier(.25,.1,.25,1)',
        }
      );

      // Once the animation is finished, permanently place the div on the screen
      // (or else it will return to it's starting offscreen location)
      animation.onfinish = ()=>{
        if (animateIn)
          this.mainDiv.style.left = String(targetLeft) + "px";
        else
          document.body.removeChild(this.mainDiv);
      };
    }
  }

  set_difference(setA, setB) {
    const _difference = new Set(setA);
    for (const elem of setB) {
      _difference.delete(elem);
    }
    return _difference;
  }

  devListChangeNotification() {
    var newDevs = new Set(this.appdata.networkDeviceList.deviceList);
    var oldDevs = new Set(this.devices);

    var toRemove = this.set_difference(oldDevs, newDevs);
    var toAdd = this.set_difference(newDevs, oldDevs);

    var myDevViews = this.deviceViews.slice();

    for (const elem of toRemove) {
      var idx = this.devices.indexOf(elem);
      if (idx > -1) {
        this.devices.splice(idx, 1);
      }
    }

    for (var elem of toAdd) {
      var idx = this.devices.indexOf(elem);
      if (idx == -1) {
        this.devices.push(elem);
      }
    }

    const sorted = this.devices.sort((x, y) => {
      const xu = x.getFriendlyName().toUpperCase();
      const yu = y.getFriendlyName().toUpperCase();
      if (xu.localeCompare(yu) < 0)
        return -1;
      else if (yu.localeCompare(xu) < 0)
        return 1;

      return x.serial.localeCompare(y.serial);
    });

    for (const elem of myDevViews) {
      elem.detach();
    }

    this.deviceViews = [];
    for (const dev of this.devices) {
      var v = new DeviceListItemView(dev);
      v.attachTo(this.innerDiv);
      this.deviceViews.push(v);
    }
  }
};


function ShowDiscoveryView(shouldAnimate, userInitiated) {
  if (shouldAnimate == undefined) {
    shouldAnimate = true;
  }

  if (userInitiated == undefined) {
    userInitiated = false;
  }

  var discoveryDiv = document.getElementById("DeviceListContainer");
  if (discoveryDiv != undefined)
    return undefined; // already showing

  var w = document.body.clientWidth;
  var h = document.body.clientHeight;

  // Slide in from the left
  var l = -w;

  // The discovery view will assign itself to an appdata property the first time it's created.
  let discovery = gApp.appdata.discoveryView ? gApp.appdata.discoveryView : new DeviceList(-w, 0, w, h, gApp.appdata);
  discovery.mainDiv.style.left = l + "px";

  if (userInitiated)
    discovery.mainDiv.setAttribute("data-userInitiated", "");
  else if (discovery.mainDiv.hasAttribute("data-userInitiated")) {
    discovery.mainDiv.removeAttribute("data-userInitiated");
  }

  const canAnimate = discovery.mainDiv.animate && typeof discovery.mainDiv.animate == "function";
  document.body.appendChild(discovery.mainDiv);

  if (shouldAnimate && canAnimate) {
    let animation = discovery.mainDiv.animate(
      [
        {
          // from
          left: discovery.mainDiv.style.left
        },
        {
          // to
          left: "0px"
        }
      ],
      {
        duration: kDiscoverySlideDuration,
        easing: 'cubic-bezier(.25,.1,.25,1)',
      }
    );

    // Once the animation is finished, permanently place the div on the screen
    // (or else it will return to it's starting offscreen location)
    animation.onfinish = ()=>{
      discovery.mainDiv.style.left = "0px";
    };
  }

  if (!userInitiated || !canAnimate) {
    discovery.mainDiv.style.left = 0;
  }
  return discovery;
}

function HideDiscoveryView(shouldAnimate) {
  if (shouldAnimate == undefined) {
    shouldAnimate = true;
  }

  let discoveryDiv = document.getElementById("DeviceListContainer");

  if (discoveryDiv && typeof discoveryDiv.getAnimations === 'function')
  {
    let animations = discoveryDiv.getAnimations();
    if (animations != undefined && animations.length > 0) {
      for (var a of animations) {
        a.cancel();
      }
    }
  }

  if (discoveryDiv != undefined) {
    const canAnimate = discoveryDiv.animate && typeof discoveryDiv.animate == "function";

    if (shouldAnimate && canAnimate) {
      var w = document.body.clientWidth;
      let animation = discoveryDiv.animate(
        [
          {
            // from
            left: discoveryDiv.style.left
          },
          {
            // to
            left: String(-w) + "px",
          }
        ],
        {
          duration: kDiscoverySlideDuration,
          easing: 'cubic-bezier(.25,.1,.25,1)',
        }
      );
      animation.onfinish = ()=> { 
        if (document.body.contains(discoveryDiv))  // Sanity check
          document.body.removeChild(discoveryDiv); 
      };
    }
    else {
      // No animation, just remove the div
      if (document.body.contains(discoveryDiv)) 
        document.body.removeChild(discoveryDiv);
    }
  }
}

