'use strict';

class DiagnosticsModalView extends View {
  constructor(appdata) {
    super(0,0,kFxWidth,kFxHeight);

    this.setTimer = this.setTimer.bind(this);
    this.onTimer = this.onTimer.bind(this);
    this.activate = this.activate.bind(this);
    this.onAEErr = this.onAEErr.bind(this)
    this.onDS = this.onDS.bind(this)
    this.wantsMouse = true;
    this.setBG(kBlackBG);

    this.modal = new ModalCanvas(appdata, false);
    this.modal.mdiv.style.zIndex = kZIndexDiagnostics; // lower than updater
    this.modal.deactivate();
    this.active = false;

    this.bufferStatusOk = true;
    this.dpcLatencyMax = 0;
    this.dpcLatencyWorst = 0;
    this.curSampleRate = 48000;
    this.curBufferSize = 256;
    this.aeErrs = {};
    this.gen5Lib = appdata.gen5Lib;

    appdata.datastore.addListener('kSampleRate', this.onDS);
    appdata.datastore.addListener('kHostBuffer1x', this.onDS);
    appdata.eventSys.addListener('kWinAEErr', this.onAEErr);

    this.okRect = new Rect(this.r - 100, this.b - 40, 80, 30);
    this.clearRect = new Rect(this.okRect.l - 120, this.okRect.t, this.okRect.w, this.okRect.h);
  }
  activate(showIt) {
    if(!showIt)
      this.modal.deactivate();
    else {
      this.modal.activate(this);
      this.requestRender();
    }
    // check the diagnostics once every 5 seconds
    this.active = showIt;
    this.setTimer(5000);
  }
  handleMouseDown(xy) {
    if(this.okRect.contains(xy.x, xy.y))
      this.modal.deactivate();
    else if(this.clearRect.contains(xy.x, xy.y)) {
      this.dpcLatencyMax = 0;
      this.dpcLatencyWorst = 0;
      this.aeErrs = {};
      this.requestRender();
    }
  }
  setTimer(ms) {
    if(this.timerID != undefined)
      clearTimeout(this.timerID);
    this.timerID = setTimeout(this.onTimer, ms);
  }
  onTimer() {
    this.timerID = undefined;
    if(this.active) {
      var d = this.gen5Lib.getDiagnostics();
      if(!d) {
        this.activate(false);
        return;
      }

      // the max time spent processing a callback from when the device sent the interrupt to when the callback exited
      if (d.minUSBLat != 64) // engine sets to 64 when it has no diagnostics
      {
          this.dpcLatencyMax = Math.max((d.maxUSBLat - 1) * 125, 125);
          var maxTime_Sec= (this.dpcLatencyMax + d.maxCBus) / 1000000.0;
          var bufferTime_Sec = this.curBufferSize / this.curSampleRate;
          this.bufferStatusOk = (1.2 * bufferTime_Sec) > maxTime_Sec;
      }

      if (this.dpcLatencyMax > this.dpcLatencyWorst)
        this.dpcLatencyWorst = this.dpcLatencyMax;

      this.setTimer(5000);
      this.requestRender(0);
    }
  }
  onAEErr(err) {
    aeErrs[err] = new Date();
    this.requestRender();
  }
  onDS(k, v, setByObject) {
    if(k == 'kSampleRate')
      this.curSampleRate = v[0];
    else if(k == 'kHostBuffer1x')
      this.curBufferSize = v[0];
  }
  render(c){
    c.fillStyle = c.strokeStyle = kToggleDefault;
    c.font = "Lighter 18px Helvetica";
    c.textAlign = 'left';

    c.lineWidth = 4;
    c.strokeRect(this.l+2, this.t+2, this.w-4, this.h-4);

    const lineHeight = 22;
    var y = (2 * lineHeight) + fillTextWrapped(c, "The information displayed here can be helpful when troubleshooting USB audio. Ignore this view unless you are experiencing frequent audio issues.", this.l+20, this.t+20, this.w-40, lineHeight);

    if(this.dpcLatencyMax != 0)
      y = (2 * lineHeight) + fillTextWrapped(c, "USB DPC Latency: " + this.dpcLatencyMax + " μs, highest measured " + this.dpcLatencyWorst + " μs" , this.l+20, y, this.w-40, lineHeight);

    for (let [e, d] of Object.entries(this.aeErrs)) {
      var codeStr = d.toString() + " Code " + e.toString(16);
      var err = e >> 16;
      var more = e & 0xffff;
      y = lineHeight + fillTextWrapped(c, codeStr, this.l+20, y, this.w-40, lineHeight);
      switch (err)
      {
      default:
        y += lineHeight;
        codeStr + "\n";
        break;
      case 1:  // (1 << 16)
        // USB stream failed to start, memory or USB setup could cause this
        // err = kFailStartUSB = 1,
        y = (2 * lineHeight) + fillTextWrapped(c, "The computer's USB controller is reporting errors and the driver is unable to perform audio. Try another USB port on a different card, or a different USB cable.", this.l+20, y, this.w-40, lineHeight);
        break;
      case 2:
        // The isoch stream failed while operating.
        // err = kFailStreamErr = 2,
        // More holds the USBStreamStatus
        // chase underflowed, ie chase was not called often enough (probably interrupts are dropped)
        // more = kUSBStreamChaseUnderrun = 1,
        // isoch input packets were not received (dropped, crc failed, or host didn't poll)
        // more = kUSBStreamInputError = 2
        if (more == 1) { // (2 << 16) & 1
          // I'm filtering out this error unless it occurs multiple times in a short amount of times
          // because it's possible these errors may happen once in a while and we don't want folks to be alarmed
          //if (aeErrTimes.ContainsKey(kv.Key) && (aeErrTimes[kv.Key] != 0) && (aeErrTimes[kv.Key] < 2000))
          //    errStr += codeStr + "System DPC latency was extremely high and has caused an audio dropout. This was most likely caused by another driver on this system. Please contact MOTU tech support if this problem persists.\n\n";
          y = (2 * lineHeight) + fillTextWrapped(c, "The driver was unable to process audio in time and audio was dropped. This was most likely caused by another driver on this system. Please contact MOTU tech support if this problem persists.", this.l+20, y, this.w-40, lineHeight);
        }
        else if (more == 2) { // (2 << 16) & 1
          y = (2 * lineHeight) + fillTextWrapped(c, "The computer's USB host controller has dropped audio. Try another USB port on a different card, or a different USB cable.", this.l+20, y, this.w-40, lineHeight);
        }
        else { // (2 << 16) & 1
          y += lineHeight;
        }
        break;
      case 3: // (3 << 16)
        // interrupt message sequence numbers mismatched (interrupt was dropped by controller), audio was resynchronized.
        // the driver tries to resync but if it happens a lot there will be obvious drop outs. May be improved by larger buffers
        // kResyncIntSeqErr = 3,
        y = (2 * lineHeight) + fillTextWrapped(c, "The computer's USB controller cannot perform at the current buffer size. Increase the buffer size to avoid audio dropouts. If the problem persists with larger buffer sizes then there may be a problem with the USB controller on this system.", this.l+20, y, this.w-40, lineHeight);
        break;
      }
    }

    // if (string.IsNullOrEmpty(errStr))
    //     diagnosticsOutput.AppendText("No diagnostic events have been reported");
    // else
    //     diagnosticsOutput.AppendText(errStr);

    drawButton(c, this.okRect, "OK", false, kToggleDefault);
    drawButton(c, this.clearRect, "Clear", false, kToggleDefault);
  }
}


class WaveEnableMultiSelect extends MultiSelect {
  constructor(d, l, t, w, h, ctrlID, appdata, dsKey) {
    super(d, l, t, w, h, ctrlID, appdata.eventSys);

    this.datastore = appdata.datastore;
    this.onDSArray = this.onDSArray.bind(this);

    SetupDSArrayListenerBasic(appdata.datastore, this, dsKey, 0);
    this.onDSArray(dsKey, DSArrayGetBasic(appdata.datastore, this));
    this.scrollBarV.setAutoHide(false);
    this.scrollBarV.setColor("rgb(20,100,155)");
  }
  setup(names) {
    this.clearOptions();
    const that = this;
    names.forEach(function(option){
      that.addOption(option.n, option.b);
    });
  }
  changeCB(id, checked){
    var v = DSArrayGetBasic(this.datastore, this);
    if(checked)
      v |= (1 << id);
    else
      v &= ~(1 << id);
    DSArraySetBasic(this.datastore, this, v);
  }
  onDSArray(k,v) {
    var n = this.cBoxes.length;
    for(var i = 0; i < n; i++) {
      this.cBoxes[i].checked = v & (1 << this.cBoxes[i].v);
    }
  }
}

class WaveEnableView extends View {
  constructor(c, l, t, appdata) {
    let topOffset = 30;
    super(l, t, 460, 205 + topOffset);

    c.addSub(this.wi = new WaveEnableMultiSelect(c.div, l, t + topOffset, 200, 150, 'wavei', appdata, 'kWaveInEnables'));
    c.addSub(this.wo = new WaveEnableMultiSelect(c.div, l + 230, t + topOffset, 200, 150, 'waveo', appdata, 'kWaveOutEnables'));
    c.addSub(this.lockSRcb = new LockWaveSRCheckbox(c.div, l + 25, t+ topOffset + 160, appdata.datastore));

    this.setBG(kEmptyBG);
  }

  setVis(v) {
    super.setVis(v);
    this.wi.setVis(v);
    this.wo.setVis(v);
    this.lockSRcb.setVis(v);
  }

  render(c) {
    if(this.wi) {
      c.fillStyle = "white";
      c.font = "300 16px Arial";
      c.textAlign = 'left';
      c.fillText("Enable Windows Audio Streams:", this.l + 2, this.wi.t - 15);
    }
  }
}

function getBufferSizeArray(datastore) {
  var sr = datastore.getKVArrayItem('kSampleRate', 0);
  var srMult = getSampleRateMult(sr);
  if (srMult == 1) {
    return kHostBuffers1x;
  }
  else if (srMult == 2) {
    return kHostBuffers2x;
  }
  return kHostBuffers4x;
}

class BufferSizeMenu extends DSLabelSelectMenu {
  constructor(d, l, t, datastore) {
    super(d, l, t, "Host Buffer Size:", datastore, 'kHostBuffer1x', getBufferSizeArray(datastore), 0);

    this.datastore = datastore;
    this.myDiv = d;
    this.srChanged = this.srChanged.bind(this);
    datastore.addListener('kSampleRate', this.srChanged);
  }

  srChanged() {
    // Choose the correct values array for the sample rate.
    var sr = this.datastore.getKVArrayItem('kSampleRate', 0);
    var srMult = getSampleRateMult(sr);
    this.setMenuVals(this.myDiv, 'kHostBuffer1x', getBufferSizeArray(this.datastore), 0); // Replace the menu object

    // change the labels of the options to match the mult
    var options = this.menu.menu.sel.options;
    var len =  len = options.length;
    for (var i = 0; i < len; i++ ) {
      options[i].text = options[i].value * srMult;
    }
  }
}

class SaveLoadPreset extends View {
  constructor(l, t, datastore) {
    super(l, t, 330, 40);

    this.datastore = datastore;
    this.savePresetR = new Rect(190, t + 5, 70, 30);
    this.loadPresetR = new Rect(290, t + 5, 70, 30);
    this.wantsMouse = true;

  }
  handleMouseDown(xy) {
    if(this.savePresetR.contains(xy.x, xy.y))
      this.datastore.savePreset();
    else if(this.loadPresetR.contains(xy.x, xy.y))
      this.datastore.loadPreset();
  }
  offset(dx,dy) {
    if (this.savePresetR)
      this.savePresetR.offset(dx,dy);
    if (this.loadPresetR)
      this.loadPresetR.offset(dx,dy);
    super.offset(dx,dy);
  }
  render(c) {
    c.fillStyle = "white";
    c.font = "300 16px Arial";
    c.textAlign = 'left';
    c.fillText("Device Presets:", this.l, this.t + 25);
    drawButton(c, this.savePresetR, "Save...", false, "black");
    drawButton(c, this.loadPresetR, "Load...", false, "black");
  }
}


class DevicePasswordView extends View {
  constructor(l, t, device) {
    super(l, t, 330, 60);

    this.render = this.render.bind(this);
    this.passwordEnabled = this.passwordEnabled.bind(this);
    this.handleMouseDown = this.handleMouseDown.bind(this);
    this.passwordEnableStatusDidChange = this.passwordEnableStatusDidChange.bind(this);

    this.device = device;
    this.clearButtonR = new Rect(190, t + 5, 70, 30);
    this.setButtonR = new Rect(290, t + 5, 70, 30);

    this.wantsMouse = true;

    this.device.eventSys.addListener("PasswordEnableStatusChanged", this.passwordEnableStatusDidChange);
  }

  passwordEnableStatusDidChange() {
    this.requestRender();
  }
  handleMouseDown(xy) {
    if(this.clearButtonR.contains(xy.x, xy.y))
    {
      this.device.resetPassword();
    }
    else if(this.setButtonR.contains(xy.x, xy.y))
    {
      gApp.showPasswordEntry(
        PasswordEntryType.SetPassword,
         "",
        // Callbacks...
        {
          canceled: ()=>{ 
            console.log("Password change aborted");
            // this.eventSys.dispatchASync("DeviceAuthenticationAborted");
            gApp.closePasswordEntry(); 
            // this.disconnect();
          },
          completion: (new_pw)=> {
            console.log("Password change complete");
            let result = this.device.submitChangePasswordRequest(new_pw);
            if (result) {
              if (this.device.netDevice != undefined) {
                (new PWCache()).savePassword(this.device.netDevice.serial, new_pw);
              }
            }
            else {
              // TODO?
              console.log("Password change failed");
            }

            gApp.closePasswordEntry();
          }
        }
      );
    }
  }

  passwordEnabled() {
    if (this.device != undefined) {
      return this.device.isPasswordEnabled;
    }
    return false;
  }

  offset(dx, dy) {
    if (this.clearButtonR)
      this.clearButtonR.offset(dx,dy);
    if (this.setButtonR)
      this.setButtonR.offset(dx,dy);
    super.offset(dx,dy);
  }

  render(c) {
    c.fillStyle = kEmptyBG;
    c.fillRect(this.l, this.t, this.w, this.h);

    c.fillStyle = "white";
    c.font = "300 16px Arial";
    c.textAlign = 'left';
    c.fillText("Device Password:", this.l, this.t + 25);

    drawButton(c, this.clearButtonR, "Clear", false, "black", this.passwordEnabled());
    drawButton(c, this.setButtonR, "Set", false, "black");
  }
}


class DiagnosticsIcon extends View {
  constructor(appdata, l, t) {
    super(l, t, 20, 20);
    this.wantsMouse = true;
    this.diagnosticsView = new DiagnosticsModalView(appdata);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y))
      this.diagnosticsView.activate(true);
  }
  render(c) {
    drawDiagnosticsIcon(c, this.cx, this.cy, "white");
  }
}

class LockWaveSRCheckbox extends CheckBox {
  constructor(d, l, t, datastore) {
    super(d, l, t, "Sync Windows sample rate to device", "waveSRCenable_id", (_, c)=>{this.changeCB(c);});
    this.waveSRCEnableChanged = this.waveSRCEnableChanged.bind(this);
    this.changeCB = this.changeCB.bind(this);

    this.datastore = datastore;
    this.wantsMouse = true;

    this.waveSRCEnableChanged();
  }

  waveSRCEnableChanged() {
    this.check(!this.datastore.getKVArrayItem('kWaveSRCEnabled', 0));
  }

  changeCB(checked){
    this.datastore.setV('kWaveSRCEnabled', 0, !checked, this);
  }
}

class DevNameView extends View {
  constructor(l,t,appdata) {
    super(l,t,500,40);

    this.deviceNameChanged = this.deviceNameChanged.bind(this);
    this.getDeviceName = this.getDeviceName.bind(this);
    this.calcBox = this.calcBox.bind(this);

    this.appdata = appdata;
    this.nameBoxL = this.l + (320 / 2);

    this.nameBox = new Rect(this.nameBoxL, t, this.w/2, 28);

    this.wantsMouse = true;
    this.enableClipping(true);

    this.setBG(kEmptyBG);

    this.appdata.datastore.addListener("kDevName", this.deviceNameChanged);
  }

  setRoot(r) {
    super.setRoot(r);
    if (r) {
      this.calcBox();
    }
  }

  calcBox() {
    const root = this.root;
    if (root) {
      const dname = this.getDeviceName();
      let c = root.context;
      c.save();
      
      c.font = "300 16px Arial";
      const strdim = stringDimensions(dname, root.context, root);

      c.restore();
      this.nameBox = new Rect(this.nameBoxL, this.t, Math.max(250, strdim.width), Math.max(28, strdim.height + 3));
      this.setBox = new Rect(this.nameBoxL, this.t + 5, 70, 30);
      this.requestRender();
    }
  }

  deviceNameChanged(k, v) {
    this.calcBox();
    this.requestRender();
  }

  handleMouseDown(xy) {
    if (this.nameBox.contains(xy.x, xy.y)) {
      let w = 400;
      let h = 250;

      let cx = document.body.clientWidth / 2.0;
      let cy = document.body.clientHeight / 2.0;

      let x = cx - (w/2.0);
      let y = cy - (h/2.0);

      const callbacks = {
        completion: (v, dialog) =>{
          if (dialog != undefined && typeof dialog.dismiss == "function")
            dialog.dismiss();
          this.appdata.datastore.setV("kDevName", 0, v);
        },

        canceled: (dialog) => {
          dialog.dismiss();
        }
      };

      new TextEntryDialog(x,y,w,h, gApp.appdata, "Enter Device Name", "", callbacks, false, this.validate);
    }
  }

  offset(dx, dy) {
    if (this.nameBox)
      this.nameBox.offset(dx, dy);
    if (this.setBox)
      this.setBox.offset(dx, dy);
    super.offset(dx,dy);
  }

  validate(str) {
    const bytes = new TextEncoder("utf-8").encode(str);
    if (bytes.length > 31) {
      return {
        ok: false,
        errMsg: "UTF-8 encoded length cannot exceed 31 bytes"
      };
    }
    else {
      return {
        ok: true,
        errMsg: undefined
      }
    }
  } 

  getDeviceName() {
    return this.appdata.datastore.getKVArrayItem("kDevName", 0);
  }

  offset(dx, dy) {
    if (this.nameBox)
      this.nameBox.offset(dx, dy);
    if (this.setBox)
      this.setBox.offset(dx, dy);
    super.offset(dx, dy);
  }

  render(c) {
    c.font = "300 16px Arial";

    c.strokeStyle = "white";
    c.fillStyle = "white";

    c.fillText("Device Name:", this.l, this.t + 25);

    let devName = this.getDeviceName();

    if (devName == "" || devName.charCodeAt(0) == 0) {
      drawButton(c, this.setBox, "Set", false, "black");
    }
    else {
      c.fillStyle = kToggleDefault;
      c.fillText(devName, this.nameBox.l, this.t + 25);
    }
  }
}

class HPaddedView extends View {
  constructor(padding, v) {
    let myRect = new Rect(v.l, v.t, v.w, Math.max(padding, v.h));
    super(myRect.l, myRect.t, myRect.w, myRect.h);

    this.getView = this.getView.bind(this);

    this.addSub(v);
    this.mySub = v;
  }

  getView() {
    return this.mySub;
  }
}


class DevSetup extends View{
  constructor(appdata) {
    super(0, 0, 0, 0);
    var xPos = kTabSelW;

    this.c = new CanvasRoot(appdata.animator);
    this.c.addSub(this);

    this.resized = this.resized.bind(this);
    this.updateDriverVersion = this.updateDriverVersion.bind(this);
    this.repositionElements = this.repositionElements.bind(this);
    this.isLocalDevice = this.isLocalDevice.bind(this);
    this.isMIDIDevice = this.isMIDIDevice.bind(this);
    this.is828 = this.is828.bind(this);

    this.appdata = appdata;
    appdata.eventSys.addListener('wResized', this.resized);
    appdata.eventSys.addListener('driverVersion', this.updateDriverVersion);

    this.scrollBarVH = new VHScrollBar(this.c.div);
    this.scrollBarVH.containerDiv.id = "devSetupContainer";
    this.scrollBarVH.containerDiv.style.position = "relative";
    this.scrollBarVH.containerDiv.style.width = "calc(100% - " + xPos + "px)";
    this.scrollBarVH.containerDiv.style.height = "100vh";
    this.scrollBarVH.containerDiv.style.float = "left";

    this.c.div.id = "devSetup";
    this.c.div.style.overflow = "auto";
    this.c.div.style.position = "relative";
    this.c.div.style.height = "100vh";
    this.c.div.style.backgroundColor = kEmptyBG;
    this.c.canvas.style.position = "static";
    this.c.div.onscroll = this.resized;

    this.tabDiv = this.scrollBarVH.containerDiv;

    this.device = appdata.device;

    const device_props = this.device.properties

    let last = undefined;
    this.c.addSub(this.rate = last = new HPaddedView(60, new DSLabelSelectMenu(this.c.div, 30, 80, "Sample Rate:", appdata.datastore, 'kSampleRate', kSampleRates, 0)));
    this.c.addSub(this.src = last = new HPaddedView(60, new DSLabelSelectMenu(this.c.div, 30, last.b, "Clock Source:", appdata.datastore, 'kClockSource', this.device.properties.kClockSources, 0)));
    this.c.addSub(this.loopback = this.opt = last = new HPaddedView(60, new DSLabelSelectMenu(this.c.div, 30, last.b, "Loopback Location:", appdata.datastore, 'kLoopbackFirst', device_props.kLoopbackOrder, 0)));
    this.c.addSub(this.opt = last = new HPaddedView(60, new DSLabelSelectMenu(this.c.div, 30, last.b, "Mix Input Meters:", appdata.datastore, 'kMixInMeterMode', kMixInputMetersMode, 0)));
    this.c.addSub(this.opt = last = new HPaddedView(60, new DSLabelSelectMenu(this.c.div, 30, last.b, "Mix USB Channels:", appdata.datastore, 'kHostOutMixMode', kHostOutputMixMode, 0)));
    this.c.addSub(this.opt = last = new HPaddedView(60, new DSLabelSelectMenu(this.c.div, 30, last.b, "MIDI Thru:", appdata.datastore, 'kMIDIThru', kMIDIThruValues, 0)));
    this.c.addSub(this.opt = this.wordClockView = last = new HPaddedView(60, new DSLabelSelectMenu(this.c.div, 30, last.b, "Word Clock:", appdata.datastore, 'kWordClockOut', kWorkClockOutValues, 0)));
    this.c.addSub(this.opt = this.optExpanderView = last = new HPaddedView(60, new DSLabelSelectMenu(this.c.div, 30, last.b, "Optical Expander:", appdata.datastore, 'kOpticalExpander', kOpticalExpanderEnableValues, 0)));
    this.c.addSub(this.opt = last = new HPaddedView(60, new DSLabelSelectMenu(this.c.div, 30, last.b, "Network Control:", appdata.datastore, 'kNetControlEnable', kNetControlEnableValues, 0)));

    this.netControlView = this.opt;
    this.c.addSub(this.opt = new HPaddedView(60, new DevicePasswordView(30, last.b, appdata.device)));
    this.devicePasswordView = this.opt;

    appdata.eventSys.addListener("PasswordEnableStatusChanged", this.c.requestRenderAll);

    this.c.addSub(this.opt = new HPaddedView(60, new DevNameView(30, this.opt.b, appdata)));

    this.devNameView = this.opt;

    last = this.opt;

    this.driverVersion = undefined;

    if(!appdata.gen5Lib.hasLib()) {
      if(appdata.device.hasHostMsgPipe() && !appdata.device.has_iOS_hostMsgPipe()) { // mac is connected
        this.c.addSub(this.saveLoad = last = new HPaddedView(60, new SaveLoadPreset(30, last.b, appdata.datastore)));
        this.c.addSub(this.upd = last = new HPaddedView(60, new UpdateView(30, last.b, appdata)));

        this.driverVersion = "";
      }
      else {
        // display the update view even though there's no host so it can show serial number
        this.c.addSub(this.upd = last = new UpdateView(30, last.b, appdata));
      }
    }
    else { // pc is connected
      this.driverVersion = "";

      this.c.addSub(this.bs = last = new HPaddedView(60, new BufferSizeMenu(this.c.div, 30, last.b, appdata.datastore)));
      this.c.addSub(this.so = last = new HPaddedView(60, new DSLabelSelectMenu(this.c.div, 30, last.b, "Safety Offsets:", appdata.datastore, 'kExtraSafetyOffset1X', kExtraSafetyOffset1X, 0)));
      this.c.addSub(this.we = last = new WaveEnableView(this.c, 30, last.b, appdata));

      this.c.addSub(this.saveLoad = last = new HPaddedView(60, new SaveLoadPreset(30, last.b, appdata.datastore)));
      this.c.addSub(this.upd = last = new UpdateView(30, last.b, appdata));
      this.c.addSub(this.diag = last = new DiagnosticsIcon(appdata, this.l + 30, last.b + 26));
    }

    this.setBG(kEmptyBG);

    appdata.eventSys.addListener('DeviceStatusChanged', this.repositionElements);
    appdata.eventSys.addListener("DeviceAuthenticated", this.repositionElements);
    appdata.datastore.addListener('kModelID', this.repositionElements);

    this.repositionElements();
  }

  isLocalDevice() {
    const dev = this.appdata.device.netDevice;
    if (dev) {
      return dev.ip && (dev.ip.startsWith("127.0.0.1") || dev.ip.toLowerCase().startsWith("localhost")) || dev.ip == "midi";
    }
    return false;
  }

  isMIDIDevice() {
    const dev = this.appdata.device.netDevice;
    if (dev) {
      return dev.ip == "midi";
    }
    return false;
  }

  is828() {
    const model_id = this.appdata.datastore.getKVArrayItem('kModelID', 0);
    if (model_id.search("828") == -1)
      return false;
    return true;
  }

  repositionElements() {

    const ApplyDeltaToSubviews = (delta, afterView) => {
      let doRepos = false;
      for (const i of this.c.subViews.items) {
        if (! doRepos) {
          if (i == afterView) {
            doRepos = true;
          }
          continue;
        }
        i.offset(0, delta);
      }
    };

    const SetSubVis = (sub, shouldBeVis, yMin, yMax) => {
      if (sub && (sub.isVis != shouldBeVis)) {
        sub.setVis(shouldBeVis);
        const t = sub.t;

        if (t < yMin)
          yMin = t;
        if (sub.b > yMax)
          yMax = sub.b;
      }
      return [yMin, yMax];
    };

    const SetHorizontalSectionVis = (subs, shouldBeVis, yMin, yMax) => {
      [yMin, yMax] = SetSubVis(subs[0], shouldBeVis, yMin, yMax);

      for (const s of subs.slice(1)) {
        SetSubVis(s, shouldBeVis, 0, 0);
      }
      return [yMin, yMax];
    };

    const SetSectionVis = (subs, shouldBeVis) => {
      let yMin = Number.MAX_SAFE_INTEGER;
      let yMax = Number.MIN_SAFE_INTEGER;

      for (const s of subs) {
        if ((typeof s == "object") && Array.isArray(s)) {
          [yMin, yMax] = SetHorizontalSectionVis(s, shouldBeVis, yMin, yMax);
        }
        else {
          [yMin, yMax] = SetSubVis(s, shouldBeVis, yMin, yMax);
        }
      }
      let yDelta = (yMax > yMin) ? yMax - yMin : 0;
      if (yDelta != 0) {
        ApplyDeltaToSubviews(
          shouldBeVis ? yDelta : -yDelta,
          subs[subs.length - 1]
        );
      }
    };

    if (this.device.netDevice == undefined)
      return;

    SetSectionVis([this.wordClockView, this.optExpanderView], this.is828());

    if (! this.isLocalDevice() || this.isMIDIDevice()) {
      // Remote device
      SetSectionVis([this.netControlView, this.devicePasswordView], false);

      if (this.appdata.gen5Lib.hasLib()) {
        SetSectionVis([this.bs, this.so, this.we], false);
      }

      this.upd.setH(100);
    }
    else {
      // Local proxy device
      SetSectionVis([this.netControlView, this.devicePasswordView], true);

      if (this.appdata.gen5Lib.hasLib()) {
        SetSectionVis([this.bs, this.so, this.we], true);
      }

      this.upd.setH(180);
    }
    this.requestRender();
  }

  recreate() {
    if (this.we && this.we.wi) {
      this.we.wi.setup(this.device.properties.waveInInfo);
    }
    if (this.we && this.we.wo) {
      this.we.wo.setup(this.device.properties.waveOutInfo);
    }
    this.src.getView().setMenuVals(this.c.div, 'kClockSource', this.device.properties.kClockSources, 0);
    this.loopback.getView().setMenuVals(this.c.div, 'kLoopbackFirst', this.device.properties.kLoopbackOrder, 0);
  }
  updateDriverVersion(ev, dv) {
    this.driverVersion = dv;
  }
  resized() {
    var w = 400;
    var h = this.upd.b;
    if(this.driverVersion)
      h += 60;

    var cs = getComputedStyle(this.c.div);
    var w = Math.max(w, parseInt(cs.width, 10)) - 1;

    this.c.setVisibleRect(0, 0, w, h);

    this.setR(this.c.r);

    this.scrollBarVH.vs.updateScroll(h);
    this.scrollBarVH.hs.updateScroll(w);
  }
  render(c) {
    drawDeviceNameHeader(c, this.l + 35, this.t, 630, this.device);

    c.font = "600 11px Arial";
    c.textAlign = 'left';
    {
      let top = this.loopback.t - 5;
      kAlertImg.render(c, 358, top);
      c.fillText("Changes to Loopback Location", 405, top + 14);
      c.fillText("will reset the USB connection.", 405, top + 31);
    }

    if (this.netControlView && !this.device.isPasswordEnabled && (this.isLocalDevice() && !this.isMIDIDevice())) {
      let top = this.netControlView.t - 5;
      kAlertImg.render(c, 356, top);
      c.fillText("Device Password is not set. Enabling Network", 405, top + 8);
      c.fillText("Control without a password will allow anyone", 405, top + 25);
      c.fillText("on your network to control your device.", 405, top + 42);
    }

    c.fillStyle = "white";
    c.font = "300 16px Arial";
    c.textAlign = 'left';

    if(this.driverVersion && this.isLocalDevice()) {
      let t= this.upd.b + 40;
      if (!this.isLocalDevice())
        t = this.upd.t + 120;

      c.fillText("Driver Version:", this.l + (this.diag != undefined ? 60 : 30), t);
      c.fillStyle = kToggleDefault;
      c.fillText(this.driverVersion, this.l + 190, t);
    }
  }
}
