'use strict';

const kSF = 0.001;
const k1OverSF = 1/kSF;
const kLogCurveCalc = Math.log(1 / kSF);
const kPiMultCurveCalc = kSF * Math.PI;
const kYScaleCurveCalc = 10 / Math.LN10;

const kiEQDSKeys = {
  kEQMode:'kiEQMode',
  kEQBypass:'kiEQBypass',
  kEQFreq:'kiEQFreq',
  kEQGain:'kiEQGain',
  kEQBW:'kiEQBW'
};
const koEQDSKeys = {
  kEQMode:'koEQMode',
  kEQBypass:'koEQBypass',
  kEQFreq:'koEQFreq',
  kEQGain:'koEQGain',
  kEQBW:'koEQBW'
};

class EQBand {
  constructor(numPoints) {
    this.nPoints = numPoints;
    this.fs = 96000; // maybe todo, set this to actual rate
    this.nyquist = this.fs/2;
    this.eqtype = 'peak';
    this.gain = 0;
    this.fc = 1000;
    this.Q = 1;
    this.f1 = 0;
    this.f2 = 0;
    this.coeff = {a0:0,a1:0,a2:0,b1:0,b2:0};
    this.curve = {x:[],y:[]};
    this.peakLoc = {x:0,y:0};
    this.f1XLoc = 0;
    this.f2XLoc = 0;
    this.hitType = 'none';
    this.enabled = false;
  }

  CheckPeakHit(x, y) {
    var controlSize = 30;
    if ((x > (this.peakLoc.x-controlSize)) &&
      (x < (this.peakLoc.x+controlSize)) &&
      (y > (this.peakLoc.y-controlSize)) &&
      (y < (this.peakLoc.y+controlSize)))
    {
      this.hitType = 'peak';
      return true;
    }
    return false;
  }

  CheckBWHit(x, y) {
    if (this.eqtype !== 'peak')
    {
      return false;
    }

    var bwSize = 20;
    if (((x > (this.f1XLoc-bwSize))  &&
       (x < (this.f1XLoc+bwSize))) ||
       ((x > (this.f2XLoc-bwSize))  &&
       (x < (this.f2XLoc+bwSize))))
    {
      this.hitType = 'bw';
      return true;
    }
    return false;
  }

  Drag(plotter, x, y) {
    if (this.hitType === 'peak') {
      this.fc = plotter.XPosToFreq(x);
      if (this.eqtype !== 'highPass')
        this.gain = plotter.YPosToGain(y);
    }
    else {
      this.SetQFromFreq(plotter.XPosToFreq(x));
    }
    this.CalculateCoeff();
    this.GenerateCurve(plotter);
  }
  SetFreq(plotter, fc) {
    if(this.fc != fc) {
      this.fc = fc;
      this.CalculateCoeff();
      this.GenerateCurve(plotter);
    }
  }
  SetMode(plotter, m) {
    if(m>3)
      return;

    // m = 3 state int - high shelf = 2, low shelf = 1, peak = 0
    // eqType = 'peak','lowShelf','highShelf',highPass
    var modes = ['peak', 'lowShelf', 'highShelf', 'highPass'];
    var eqtype  = modes[m];
    if(this.eqtype != eqtype) {
      this.eqtype = eqtype;
      this.CalculateCoeff();
      this.GenerateCurve(plotter);
    }
  }
  SetGain(plotter, g) {
    if(this.gain != g) {
      this.gain = g;
      this.CalculateCoeff();
      this.GenerateCurve(plotter);
    }
  }
  SetQ(plotter, q) {
    if(this.Q != q) {
      this.Q = q;
      this.CalculateCoeff();
      this.GenerateCurve(plotter);
    }
  }

  CalculateCoeff() {
    var norm;
    var V = Math.pow(10, Math.abs(this.gain) / 20);
    var K = Math.tan(Math.PI * this.fc / this.fs);
    var Q = this.Q;
    const kQ = Math.SQRT2/2;
    switch(this.eqtype)
    {
      case 'highPass':
        norm = 1 / (1 + K / kQ + K * K);
        this.coeff.a0 = 1 * norm;
        this.coeff.a1 = -2 * this.coeff.a0;
        this.coeff.a2 = this.coeff.a0;
        this.coeff.b1 = 2 * (K * K - 1) * norm;
        this.coeff.b2 = (1 - K / kQ + K * K) * norm;
      break;
      case 'peak':
        if (this.gain >= 0) {
          norm = 1 / (1 + 1/Q * K + K * K);
          this.coeff.a0 = (1 + V/Q * K + K * K) * norm;
          this.coeff.a1 = 2 * (K * K - 1) * norm;
          this.coeff.a2 = (1 - V/Q * K + K * K) * norm;
          this.coeff.b1 = this.coeff.a1;
          this.coeff.b2 = (1 - 1/Q * K + K * K) * norm;
        }
        else {
          norm = 1 / (1 + V/Q * K + K * K);
          this.coeff.a0 = (1 + 1/Q * K + K * K) * norm;
          this.coeff.a1 = 2 * (K * K - 1) * norm;
          this.coeff.a2 = (1 - 1/Q * K + K * K) * norm;
          this.coeff.b1 = this.coeff.a1;
          this.coeff.b2 = (1 - V/Q * K + K * K) * norm;
        }
        break;
      case 'lowShelf':
        if (this.gain >= 0) {
          norm = 1 / (1 + Math.SQRT2 * K + K * K);
          this.coeff.a0 = (1 + Math.sqrt(2*V) * K + V * K * K) * norm;
          this.coeff.a1 = 2 * (V * K * K - 1) * norm;
          this.coeff.a2 = (1 - Math.sqrt(2*V) * K + V * K * K) * norm;
          this.coeff.b1 = 2 * (K * K - 1) * norm;
          this.coeff.b2 = (1 - Math.SQRT2 * K + K * K) * norm;
        }
        else {
          norm = 1 / (1 + Math.sqrt(2*V) * K + V * K * K);
          this.coeff.a0 = (1 + Math.SQRT2 * K + K * K) * norm;
          this.coeff.a1 = 2 * (K * K - 1) * norm;
          this.coeff.a2 = (1 - Math.SQRT2 * K + K * K) * norm;
          this.coeff.b1 = 2 * (V * K * K - 1) * norm;
          this.coeff.b2 = (1 - Math.sqrt(2*V) * K + V * K * K) * norm;
        }
        break;
      case 'highShelf':
        if (this.gain >= 0) {
          norm = 1 / (1 + Math.SQRT2 * K + K * K);
          this.coeff.a0 = (V + Math.sqrt(2*V) * K + K * K) * norm;
          this.coeff.a1 = 2 * (K * K - V) * norm;
          this.coeff.a2 = (V - Math.sqrt(2*V) * K + K * K) * norm;
          this.coeff.b1 = 2 * (K * K - 1) * norm;
          this.coeff.b2 = (1 - Math.SQRT2 * K + K * K) * norm;
        }
        else {
          norm = 1 / (V + Math.sqrt(2*V) * K + K * K);
          this.coeff.a0 = (1 + Math.SQRT2 * K + K * K) * norm;
          this.coeff.a1 = 2 * (K * K - 1) * norm;
          this.coeff.a2 = (1 - Math.SQRT2 * K + K * K) * norm;
          this.coeff.b1 = 2 * (K * K - V) * norm;
          this.coeff.b2 = (V - Math.sqrt(2*V) * K + K * K) * norm;
        }
        break;
    }
  };

  GenerateCurve(plotter) {
    this.curve = {x:[],y:[]};
    var a0 = this.coeff.a0, a1 = this.coeff.a1, a2 = this.coeff.a2;
    var b1 = this.coeff.b1, b2 = this.coeff.b2;
    const kPowa012 = Math.pow(a0+a1+a2, 2);
    const k4multa012 = 4*(a0*a1 + 4*a0*a2 + a1*a2);
    const k16a02 = 16*a0*a2;
    const kPowb12 = Math.pow(1+b1+b2, 2);
    const k4multb12 = 4*(b1 + 4*b2 + b1*b2);
    const k16b2 = 16*b2;
    const l = plotter.r.l;

    for(var x = 0; x < this.nPoints; x++) {
      // kSF to 1, times pi, log scale
      //const w = Math.exp(kLogCurveCalc * x / (this.nPoints - 1)) * kPiMultCurveCalc;
      const f = plotter.XPosToFreqUnbounded(x + l) / (this.nyquist);
      const w = f * Math.PI;
      const phi = Math.pow(Math.sin(w/2), 2);
      const phi2 = phi*phi;
      var y = kYScaleCurveCalc *
        (Math.log(kPowa012 - k4multa012 * phi + k16a02 * phi2) -
        Math.log(kPowb12 - k4multb12 * phi + k16b2*phi2));
      if (y == -Infinity)
        y = -100;
      // curve is not in any views coordinate space
      this.curve.y.push(y);
      this.curve.x.push(x);
    }
    this.GenerateQBWFreqs();
    this.peakLoc = {x:plotter.FreqToXPos(this.fc),y:plotter.GainToYPos((this.eqtype == 'highPass') ? 0 : this.gain)};
    this.f1XLoc = plotter.FreqToXPos(this.f1);
    this.f2XLoc = plotter.FreqToXPos(this.f2);
  };

  GenerateQBWFreqs() {
    this.f1 = this.fc * (Math.sqrt(1 + 1/(4*this.Q*this.Q)) - 1/(2*this.Q));
    this.f2 = this.fc * (Math.sqrt(1 + 1/(4*this.Q*this.Q)) + 1/(2*this.Q));
  }

  SetQFromFreq(freq) {
    var Q;

    if (freq > this.fc)
      Q = this.fc / (freq - this.fc*this.fc/freq);
    else
      Q = this.fc / (this.fc*this.fc/freq - freq);
    if (Q < 0.45)
      Q = 0.45;
    if (Q > 10)
      Q = 10;

    this.Q = Q;
  }
}


class Plotter {
  constructor(r) {
    this.r = r;
    this.maxGain = 20;
    this.fs = 48000;
    this.nyquist = this.fs/2;

    this.gridFreqs = [30, 40, 50, 60, 70, 80, 100, 200, 400, 600, 800, 1000, 2000, 3000, 4000, 5000, 6000, 7000, 8000, 10000, 20000];
    this.gridMags = [0, 6, 12, 18];

    this.yGrid = [];
    this.xGrid = [];

    this.GenerateGridData();
  }
  YPosToGain(y) {
    return (1 - (2 * (y - this.r.t))/this.r.h) * this.maxGain;
  };

  XPosToFreqUnbounded(x) {
    return (kSF*this.nyquist) * Math.exp(Math.log(k1OverSF) * (x - this.r.l)/(this.r.w - 1));
  };

  XPosToFreq(x) {
    var freq = this.XPosToFreqUnbounded(x);
    if (freq > 20000){
      return 20000;
    }
    if (freq < 25){
      return 25;
    }
    return freq;
  };

  FreqToXPos(freq) {
    var width = this.r.w;
    var fs = this.fs;

    var f2x = function(f) {
      return Math.log(2.0*f/(kSF*fs))/Math.log(k1OverSF) * (width - 1)
    };

    if (freq == undefined)
    {
      return;
    }
    if (freq.constructor === Array) {
      var x = [];
      for (var i = 0; i < freq.length; i++) {
        x.push(this.r.l + f2x(freq[i]));
      }
      return x;
    }

    return this.r.l + f2x(freq);
  };

  GainToYPos(gain) {
    var height = this.r.h;
    var maxGain = this.maxGain;

    var g2y = function(g) {
      if (g > 0)
        return height/2 - (g/maxGain * height/2);
      else
        return height/2 + (-g/maxGain * height/2);
    };

    if (gain.constructor === Array) {
      var y = [];
      for (var i = 0; i < gain.length; i++) {
        y.push(this.r.t + g2y(gain[i]));
      }
      return y;
    }

    return this.r.t + g2y(gain);
  };

  GenerateGridData() {
    this.xGrid = this.FreqToXPos(this.gridFreqs);
    this.yGrid = this.GainToYPos(this.gridMags);

    for(var m = 0; m < this.gridMags.length; m++) {
      if (this.gridMags[m])
        this.yGrid.push(this.GainToYPos(-this.gridMags[m]));
    }
  };

  DrawGrid(c) {
    c.lineWidth = 1;
    c.strokeStyle = kGridColor;

    c.beginPath();
    c.strokeRect(this.r.l+1, this.r.t+1, this.r.w-2, this.r.h-2);

    for(var x = 0; x < this.xGrid.length; x++) {
      c.moveTo(this.xGrid[x], this.r.t);
      c.lineTo(this.xGrid[x], this.r.b);
    }

    for(var y = 0; y < this.yGrid.length; y++) {
      c.moveTo(this.r.l, this.yGrid[y]);
      c.lineTo(this.r.r, this.yGrid[y]);
    }
    c.stroke();
  };

  DrawGridLabels(c) {
    c.fillStyle = kFxEditBG;
    c.rect(this.r.l, this.r.b, this.r.w, 18);
    c.fill();

    c.font = "bold 14px Arial";
    c.fillStyle = kGridColor;
    c.textAlign = 'center';

    for(var x = 0; x < this.xGrid.length; x+=2) {
      var fLabel = '';
      if(this.gridFreqs[x] >= 1000)
        fLabel = String(this.gridFreqs[x]/1000) + 'k';
      else
        fLabel = String(this.gridFreqs[x]);
      c.fillText(fLabel, this.xGrid[x], this.r.b + 12);
    }
  }

  // y: pixels
  FreqPlot(c, data, color, width) {
    if (data.x !== undefined) {
      var x = this.FreqToXPos(data.x);
      c.beginPath();
      c.moveTo(x[0], data.y[0]);

      for(var i = 1; i < data.y.length; i++)
        c.lineTo(x[i],data.y[i]);

      c.lineWidth = width;
      c.strokeStyle = color;
      c.stroke();
    }
    else {
      this.Plot(c, data,color,width);
    }
  }

  // y: magnitude (dB) NOT pixels
  Plot(c, data, color, width, fill) {
    // note data is not in the rect coordinate space
    var hasX = data.x !== undefined ? true : false;
    var x0 = hasX ? data.x[0] : 0;

    c.beginPath();
    if ((fill !== undefined) && (x0 === 0))
      c.moveTo(this.r.l, this.GainToYPos(0));

    c.lineTo(this.r.l + x0, this.GainToYPos(data.y[0]));

    var n = data.y.length;
    if(n > 0) {
      n--;
      for(var i = 1; i <= n; i++)
        c.lineTo(this.r.l + (hasX ? data.x[i] : i), this.GainToYPos(data.y[i]));
      c.lineTo(this.r.l + (hasX ? data.x[n] : n), this.GainToYPos(data.y[n]));
    }
    c.lineWidth = width;
    if (fill !== undefined) {
      if(data.y[data.y.length - 1] !== 0)
        c.lineTo(this.r.r, this.GainToYPos(0));

      c.closePath
      c.fillStyle = fill;
      c.fill();
    }
    else {
      c.strokeStyle = color;
      c.stroke();
    }
  };
}


const kEQCtrlInfo = [
  {l:"Freq:", d:0, u:" Hz", min:20, max:20000},
  {l:"Gain:", d:1, u:" dB", min:-20, max:20},
  {l:"Q:", d:2, u:"", min:0.45, max:10},
];
function validateEQ(ctrlInfo, v) {
  v = Number(v.replace(/[^0-9\.\-]/g, '')).toFixed(ctrlInfo.d);
  v = Math.min(ctrlInfo.max, v);
  return Math.max(ctrlInfo.min, v);
}

class EQEditCtrlTextEdit extends EditableTextView {
  constructor(l,t,w,h, appdata, ctrlInfo, k) {
    super(l,t,w,h, 'eq-input-detail', '17px sans-serif', appdata.keyFilter);
    this.setup = this.setup.bind(this);
    this.onDSArray = this.onDSArray.bind(this);
    this.ctrlInfo = ctrlInfo;
    this.setBG(kFxEditBG);
    this.c = kEQValue;

    this.datastore = appdata.datastore;
    SetupDSArrayListenerBasic(this.datastore, this, k, 0);
  };
  setup(i) {
    SetDSArrayIndexBasic(this, i);
    this.v = trimFloat(DSArrayGetBasic(this.datastore, this), this.ctrlInfo.d) + this.ctrlInfo.u;
    this.requestRender();
  }
  onDSArray(k, v) {
    this.v = trimFloat(v, this.ctrlInfo.d).toString() + this.ctrlInfo.u;
    this.requestRender();
  }
  valueChanged(v) {
    v = validateEQ(this.ctrlInfo, v);
    DSArraySetBasic(this.datastore, this, v);
    return v + this.ctrlInfo.u;
  }
}


class EQEditCtrl extends View {
  constructor(l,t,w,h, appdata, ctrlInfo, k) {
    super(l,t,w,h);
    this.setBG(kFxEditBG);
    this.label = ctrlInfo.l;

    this.addSub(this.edit = new EQEditCtrlTextEdit(this.cx - 25, this.t + 5, 25 + this.w/2, this.h, appdata, ctrlInfo, k));
  }
  enable(e) {
    this.edit.enable(e);
  }
  setup(i) {
    this.edit.remove();
    this.edit.setup(i);
  }
  render(c) {
    c.fillStyle = kEQLabel;
    c.font = "18px sans-serif";
    c.textAlign = 'right';
    c.fillText(this.label, this.cx - 25, this.b);
  }
}


class EQEnableBtn extends SimpleDSBtn {
  constructor(l, t, c, datastore, dsKey, i)  {
    super(l, t, 125, 25, datastore, dsKey, i);
    this.c = c;
    this.cx = this.l + 12;
    this.cy = this.t + 12;
    this.setBG(kFxEditBG);
  }
  render(c) {

    drawPwrIcon(c, 3, this.cx, this.cy, 12, (this.v == 0) ? this.c : kEQLabel);

    c.strokeStyle = this.c;
    c.beginPath();
    c.moveTo(this.l + 40, this.cy);
    c.lineTo(this.r, this.cy);
    c.lineWidth = 5;
    c.lineCap = "round";
    c.stroke();
    c.lineCap = "butt";
  }
}


const kBand1EQModes = [0, 1, 3];
const kBand4EQModes = [0, 2];
class EQModeRadioBtn extends SimpleDSView {
  constructor(l,t,w,h, datastore, dsKey, i, modes, color)  {
    super(l,t,w,h, datastore, dsKey, i);

    this.wantsMouse = true;
    this.dragging = false;
    this.modes = modes;
    this.bw = this.w/modes.length;
    this.v = this.modes[0];
    this.color = color;
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y)) {
      var v = this.modes[Math.floor((xy.x - this.l)/this.bw)];
      if(this.v != v) {
        this.v = v;
        DSArraySetBasic(this.datastore, this, this.v);
        this.requestRender();
      }
    }
  }
  //4 state int - highPass = 3, highShelf = 2, lowShelf = 1, peak = 0
  render(c) {
    c.lineWidth = 1.2;

    var n = this.modes.length;
    var cx = this.l + this.bw/2;
    var cy = this.cy;
    for(var i = 0; i < n; ++i) {
      c.strokeStyle = this.v == this.modes[i] ? this.color : kEQLabel;
      c.beginPath();
      if(this.modes[i] == 0) { // peak
        c.moveTo(cx-10, cy);
        const co = 2;
        const s = 12;
        c.bezierCurveTo(cx-co, cy, cx-co, cy-s, cx, cy-s);
        c.bezierCurveTo(cx+co, cy-s, cx+co, cy, cx+s, cy);
        c.bezierCurveTo(cx+co, cy, cx+co, cy+s, cx, cy+s);
        c.bezierCurveTo(cx-co, cy+s, cx-co, cy, cx-s, cy);
      }
      else if(this.modes[i] == 1) { // low-shelf
        c.moveTo(cx-10, cy+10);
        c.bezierCurveTo(cx, cy+10, cx, cy, cx+10, cy);
        c.bezierCurveTo(cx, cy, cx, cy-10, cx-10, cy-10);
      }
      else if(this.modes[i] == 2) { // high-shelf
        c.moveTo(cx+10, cy-10);
        c.bezierCurveTo(cx, cy-10, cx, cy, cx-10, cy);
        c.bezierCurveTo(cx, cy, cx, cy+10, cx+10, cy+10);
      }
      else if(this.modes[i] == 3) { // high-pass
        c.moveTo(cx-10, cy+10);
        c.bezierCurveTo(cx-10, cy+5, cx-10, cy-10, cx+10, cy-10);
      }
      c.stroke();
      cx += this.bw;
    }
  }
}
class EQView extends View {
  constructor(r, appdata, isInEQ, isThumb) {
    super(r.l,r.t,r.w,r.h);

    this.setup = this.setup.bind(this);
    this.onDS = this.onDS.bind(this);

    this.datastore = appdata.datastore;
    this.isThumb = isThumb;
    this.wantsMouse = !isThumb;
    this.dragging = -1;
    this.setBG(kFxEditBG);

    this.keys = isInEQ ? kiEQDSKeys : koEQDSKeys;

    if(isThumb)
      this.gr = new Rect(this.l, this.t, this.w, this.h);
    else
      this.gr = new Rect(this.l, this.t, this.w, this.h - 200);
    this.p = new Plotter(this.gr);

    this.editCtls= []
    var eq = [];
    var l = this.l;

    const lastBand = isInEQ ? 3 : 2;
    const xInc = isInEQ ? 227 : 275;
    l += isInEQ ? 55 : 117;
    for (var n = 0; n <= lastBand; n++) {
      eq.push(new EQBand(this.p.r.w));
      eq[n].CalculateCoeff();
      eq[n].GenerateCurve(this.p);

      if(!isThumb) {
        var bCtrls = {};
        this.addSub(bCtrls.eb = new EQEnableBtn(l, this.b - 173, kEQColors[n], this.datastore, this.keys.kEQBypass, 0));
        this.addSub(bCtrls.f = new EQEditCtrl(l, this.b - 141, 140, 20, appdata, kEQCtrlInfo[0], this.keys.kEQFreq));
        this.addSub(bCtrls.g = new EQEditCtrl(l, this.b - 114, 140, 20, appdata, kEQCtrlInfo[1], this.keys.kEQGain));
        this.addSub(bCtrls.q = new EQEditCtrl(l, this.b - 87, 140, 20, appdata, kEQCtrlInfo[2], this.keys.kEQBW));
        if(n == 0)
          this.addSub(bCtrls.t = new EQModeRadioBtn(l, this.b - 48, 125, 30, this.datastore, this.keys.kEQMode, 0, kBand1EQModes, kEQColors[0]));
        else if(n == lastBand)
          this.addSub(bCtrls.t = new EQModeRadioBtn(l+33, this.b - 48, 83, 30, this.datastore, this.keys.kEQMode, 3, kBand4EQModes, kEQColors[3]));
        l += xInc;
        this.editCtls.push(bCtrls);
      }
    }
    this.eq = eq;

    this.datastore.addListener(this.keys.kEQMode, this.onDS);
    this.datastore.addListener(this.keys.kEQBypass, this.onDS);
    this.datastore.addListener(this.keys.kEQFreq, this.onDS);
    this.datastore.addListener(this.keys.kEQGain, this.onDS);
    this.datastore.addListener(this.keys.kEQBW, this.onDS);
  }
  offset(dx, dy) {
    super.offset(dx, dy);
    this.gr.offset(dx,dy); // shared by plotter
  }
  onDS(k, v) {
    if(this.info != undefined) {
      var n = this.eq.length;
      if(this.info.eq >= 0) {
        var offset = this.info.eq * n;
        switch(k) {
          case this.keys.kEQBypass: {
            for(var i = 0; i < n; ++i)
              this.eq[i].enabled = !v[offset + i];
          } break;
          case this.keys.kEQMode: {
            for(var i = 0; i < n; ++i) {
              var mode = v[offset + i];
              this.eq[i].SetMode(this.p, mode);
              var bCtrls = this.editCtls[i];
              if(bCtrls != undefined) {
                bCtrls.g.enable(mode != 3);
                bCtrls.q.enable(mode == 0);
              }
            }
          } break;
          case this.keys.kEQFreq: {
            for(var i = 0; i < n; ++i)
              this.eq[i].SetFreq(this.p, v[offset + i]);
          } break;
          case this.keys.kEQGain: {
            for(var i = 0; i < n; ++i)
              this.eq[i].SetGain(this.p, v[offset + i]);
          } break;
          case this.keys.kEQBW: {
            for(var i = 0; i < n; ++i)
              this.eq[i].SetQ(this.p, v[offset + i]);
          } break;
        }
        this.requestRender();
      }
    }
  }
  setup(info) {
    this.info = info;
    if(info === undefined || info.eq == -1)
        return;

    var bypasss = this.datastore.getKVArray(this.keys.kEQBypass);
    var modes = this.datastore.getKVArray(this.keys.kEQMode);
    var freqs = this.datastore.getKVArray(this.keys.kEQFreq);
    var gains = this.datastore.getKVArray(this.keys.kEQGain);
    var qs = this.datastore.getKVArray(this.keys.kEQBW);

    var n = this.eq.length;
    var offset = info.eq * n;
    for(var i = 0; i < n; ++i) {
      this.eq[i].enabled = !bypasss[offset + i];
      var mode = modes[offset + i];
      this.eq[i].SetMode(this.p, mode);
      this.eq[i].SetFreq(this.p, freqs[offset + i]);
      this.eq[i].SetGain(this.p, gains[offset + i]);
      this.eq[i].SetQ(this.p, qs[offset + i]);
      var bCtrls = this.editCtls[i];
      if(bCtrls != undefined) {
        bCtrls.g.enable(mode != 3);
        bCtrls.q.enable(mode == 0);
        bCtrls.eb.setup(offset + i);
        bCtrls.f.setup(offset + i);
        bCtrls.g.setup(offset + i);
        bCtrls.q.setup(offset + i);
        if(bCtrls.t)
          bCtrls.t.setup(offset + i);
      }
    }
    this.requestRender();
  }
  handleMouseDown(xy, e) {
    if (("activeElement" in document) && (document.activeElement != e.srcElement))
      document.activeElement.blur();
    if(this.gr.contains(xy.x, xy.y)) {

      var eq = this.eq;
      var i, peakHit = false;
      // check hit in the reverse of the drawing order
      for (i = eq.length - 1; i >= 0; i--) {
        if(eq[i].CheckPeakHit(xy.x, xy.y)) {
          //pauseEvent(e);
          this.dragging = i;
          //eq[i].Drag(this.p, xy.x, xy.y);
          peakHit = true;
          this.requestRender();
          break;
        }
      }
      if(!peakHit) {
        for (i = eq.length - 1; i >= 0; i--) {
          if(eq[i].CheckBWHit(xy.x, xy.y)) {
            //pauseEvent(e);
            this.dragging = i;
            eq[i].Drag(this.p, xy.x, xy.y);
            this.requestRender();
            break;
          }
        }
      }

    }
  }
  handleMouseMoved(xy, delta) {
    if (this.dragging >= 0) {
      var x = Math.min(this.gr.r, Math.max(this.gr.l, xy.x));
      var y = Math.min(this.gr.b, Math.max(this.gr.t, xy.y));
      var eq = this.eq;
      var offset = this.info.eq * eq.length;
      eq[this.dragging].Drag(this.p, x, y);
      this.datastore.setV(this.keys.kEQFreq, this.dragging + offset, validateEQ(kEQCtrlInfo[0], eq[this.dragging].fc.toFixed(0)));
      this.datastore.setV(this.keys.kEQGain, this.dragging + offset, validateEQ(kEQCtrlInfo[1], eq[this.dragging].gain.toFixed(1)));
      this.datastore.setV(this.keys.kEQBW, this.dragging + offset, validateEQ(kEQCtrlInfo[2], eq[this.dragging].Q.toFixed(2)));
      this.requestRender();
    }
  }
  handleMouseUp() {
    this.dragging = -1;
  }
  render(c) {
    c.save();
    c.beginPath();

    c.fillStyle = kBlackBG;
    c.fillRect(this.gr.l, this.gr.t, this.gr.w, this.gr.h);

    var p = this.p;

    if(!this.isThumb) {
      p.DrawGrid(c);
      p.DrawGridLabels(c);
    }

    c.beginPath();
    c.rect(this.gr.l, this.gr.t, this.gr.w, this.gr.h);
    c.clip();
    c.beginPath();

    var sum = [];
    var eq = this.eq;
    for(var i = 0; i < eq[0].curve.y.length; i++) {
      var tmp = 0;
      for(var j = 0; j < eq.length; j++) {
        if(eq[j].enabled)
          tmp += eq[j].curve.y[i];
      }
      sum.push(tmp);
    }

    for(var i = 0; i < eq.length; i ++) {
      if(eq[i].enabled)
        p.Plot(c, eq[i].curve, kEQColors[i], 0.8, kEQCurveColors[i]);
    }
    p.Plot(c, {y:sum}, 'rgba(255, 255, 255, .6)', 2);

    if(!this.isThumb) {
      for(var i = 0; i < eq.length; i ++) {
        if(eq[i].enabled && (eq[i].eqtype == 'peak'))
        {
          p.FreqPlot(c, {x:[eq[i].f1,eq[i].f1],y:[p.r.t,p.r.b]}, kEQColors[i], 2);
          p.FreqPlot(c, {x:[eq[i].f2,eq[i].f2],y:[p.r.t,p.r.b]}, kEQColors[i], 2);
        }
      }
      for(var i = 0; i < eq.length; i ++) {
          fillCircle(c, eq[i].peakLoc.x, eq[i].peakLoc.y, 20, kEQControlGray);
          fillCircle(c, eq[i].peakLoc.x, eq[i].peakLoc.y, 8, kEQColors[i]);
      }

    }

    c.restore();
  }
}


class EQFxView  extends View {
  constructor(l,t,w,h,appdata,isInEQ) {
    super(l,t,w,h);
    this.setup = this.setup.bind(this);
    this.eqGV = new EQView(this, appdata, isInEQ, false);
    this.setBG(kFxEditBG);
  }
  setup(info) {
    this.removeSub(this.eqGV);
    this.eqGV.setup(info);
    if(info.eq != undefined && info.eq != -1) {
      this.addSub(this.eqGV);
    }
  }
}


class EQThumb  extends FxThumb {
  constructor(l, t, appdata, isInEQ, isFxChStrip, modalDialog) {
    super(l, t, isInEQ ? kEQFxType : kBusEQFxType, isFxChStrip, modalDialog);
    this.addSub(this.eqGV = new EQView(this, appdata, isInEQ, true));
  }
  setup(info) {
    super.setup(info);
    this.setValid(info !== undefined && info.eq != undefined && info.eq != -1);
    this.setVis(this.isValid);
    this.eqGV.setup(info);
  }
}


