"use strict";

function PlatformSupportsWebMIDI() {
  return navigator != undefined && navigator.requestMIDIAccess != undefined && typeof navigator.requestMIDIAccess == "function";
}

const SysexRequestBits = {
  kRequestDirectionBit: 6,
  kRequestDirectionValueHostToDevice: 0,
  kRequestDirectionValueDeviceToHost: 1,
};

const kRequestDirectionHostToDevice = SysexRequestBits.kRequestDirectionValueHostToDevice << SysexRequestBits.kRequestDirectionBit;
const kRequestDirectionDeviceToHost = SysexRequestBits.kRequestDirectionValueDeviceToHost << SysexRequestBits.kRequestDirectionBit;
const kRequestDirectionMask = 1 << SysexRequestBits.kRequestDirectionBit;
const kRequestID_Mask = (~kRequestDirectionMask) & 0x7f;


const kDeviceSysexProtocolID =  Uint8Array.from([0x0, 0x1]);
const kMOTUSysexHeaderBytes = Uint8Array.from([0xF0, 0x0, 0x0, 0x3b, kDeviceSysexProtocolID[0], kDeviceSysexProtocolID[1]]);
const kMOTUSysexHeader = Array.from(kMOTUSysexHeaderBytes);

const SysexRequestIDs = {
  kSetProperty: 0,
  kProtocolProbe: 1,
  kEnableSysexAPI : 2
};

function ConstructRequestValue(direction, id) {
  return direction | (id & kRequestID_Mask);
}

function IsMOTU_Msg(buf) {
  if (buf.length < 5)
      return false;

  for (var i = 0; i < kMOTUSysexHeader.length; ++i) {
      if (buf[i] != kMOTUSysexHeader[i])
          return false;
  }
  return true;
}
function GetMsgType(buf) {
  if (buf.length < kMOTUSysexHeader.length + 1)
      return undefined;
  let dir = (buf[6] & kRequestDirectionMask); 
  if (dir != kRequestDirectionDeviceToHost)
      return undefined;
  return buf[6] & kRequestID_Mask;
}

function DecodeSetPropertySysexMessage(data) {

  if (data.length < 10) {
    ProxyLog("Sysex too short");
    return undefined;
  }
  if (data[0] != 0xF0) {
    ProxyLog("Message is not Sysex");
    return undefined;
  }

  const b1 = data[1];
  const b2 = data[2];
  const b3 = data[3];

  if (b1 != 0 || b2 != 0 || b3 != 0x3b) {
    ProxyLog("Unexpected sysex header");
    return undefined;
  }

  if (data[4] != kDeviceSysexProtocolID[0] || data[5] != kDeviceSysexProtocolID[1]) {
    // ProxyLog("Unexpected sysex protocol");
    return;
  }

  let decoded = new Uint8Array(data.length);

  let i = 0;
  let n = 7;
  let shl = 7;
  let b7 = data[n++];

  while (data[n] != 0xF7) {
    if (shl == 0) {
      b7 = data[n++];
      shl = 7;
    }
    else {
      decoded[i++] = ((b7 << shl--) & 0x80) | data[n++];
    }
  }

  let len = i - 4;

  // The "decoded" buffer might be too big.
  // Allocate a properly-sized buffer to send to the
  // websocket-centric code.
  let result = new Uint8Array(len+4);
  for (var j = 0; j < len + 4; ++j) {
    result[j] = decoded[j];
  }

  return result;
}

// I'm referencing DeviceProps_828 here, but
// these properties have the same ids for the ULmk5
const WebMIDI_RequiredProps = Object.freeze(
  [
    DeviceProps_828.defaultDatastore.kModelID.id, 
    DeviceProps_828.defaultDatastore.kAPIVersion.id,
    DeviceProps_828.defaultDatastore.kDevName.id,
    DeviceProps_828.defaultDatastore.kDevUID.id
  ]
);

class WebMIDI_Device {
  constructor(input, output, motuDeviceCallback) {
    this.rxMIDI = this.rxMIDI.bind(this);
    this.sendMIDI = this.sendMIDI.bind(this);
    this.probeSysexProtocol = this.probeSysexProtocol.bind(this);
    this.enableSysexPropertyAPI = this.enableSysexPropertyAPI.bind(this);
    this.didBufferAllRequiredProps = this.didBufferAllRequiredProps.bind(this);
    this.pushOutBufferedProps = this.pushOutBufferedProps.bind(this);
    this.rxPropertyChange = this.rxPropertyChange.bind(this);
    this.massageSerialNumber = this.massageSerialNumber.bind(this);
    this.parseRequiredProp = this.parseRequiredProp.bind(this);

    this.input = input;
    this.output = output;
    this.isMOTU = false;
    this.propertyDelegate = undefined;
    this.motuCallback = motuDeviceCallback;
    this.requiredProps = new Map();
    this.didSync = false;

    input.onmidimessage = this.rxMIDI;
    output.open().then(()=>{
      this.probeSysexProtocol();
    });
  }

  didBufferAllRequiredProps() {
    for (const p of WebMIDI_RequiredProps) {
      if (!this.requiredProps.has(p))
        return false;
    }
    return true;
  }

  pushOutBufferedProps() {
    this.massageSerialNumber();

    if (this.isMOTU && this.motuCallback) {
      this.motuCallback(this);
      this.motuCallback = undefined;
    }

    if (!this.propertyDelegate)
      return;

    if (this.requiredProps.has(DeviceProps_828.defaultDatastore.kModelID.id)) {
      const buf = this.requiredProps.get(DeviceProps_828.defaultDatastore.kModelID.id);
      this.propertyDelegate.decodeSetPropertyMessage(buf);
    }

    this.requiredProps.forEach((value, key)=>{
      if (key != DeviceProps_828.defaultDatastore.kModelID.id) {
        this.propertyDelegate.decodeSetPropertyMessage(value);
      }
    });
    this.didSync = true;
  }

  massageSerialNumber() {
    if (!this.serial || this.serial.length == 0) {
      if (this.modelID && this.modelID.length == 4) {
        this.serial = (this.modelID + "FFFFFF").toUpperCase();
      }
    }
  }

  parseRequiredProp(idx, decoded) {
    if (idx == DeviceProps_828.defaultDatastore.kDevUID.id) {
      const nullTerminated = new Uint8Array(11);
      const x = decoded.slice(4);
      if (x.length != 10) {
        console.log("Bad device sn!");
        return;
      }
      for (let i = 0; i < 10; ++i) {
        nullTerminated[i] = x[i];
      }
      nullTerminated[10] = 0;
      this.serial = new TextDecoder('utf-8').decode(nullTerminated);
    }
    else if (idx == DeviceProps_828.defaultDatastore.kModelID.id) {
      this.modelID = new TextDecoder('utf-8').decode(decoded.slice(4, decoded.indexOf(0, 4)));
    }
    else if (idx == DeviceProps_828.defaultDatastore.kDevName.id) {
      this.deviceName = new TextDecoder('utf-8').decode(decoded.slice(4, decoded.indexOf(0, 4)));
    }
  }

  rxPropertyChange(buf) {
    if (!this.didBufferAllRequiredProps()) {
      const decoded = DecodeSetPropertySysexMessage(buf);
      if (decoded && decoded.length > 2) {
        const idx = (decoded[0] << 8) | (decoded[1]);
        if (WebMIDI_RequiredProps.indexOf(idx) != -1) {
          this.requiredProps.set(idx, buf);
          this.parseRequiredProp(idx, decoded);

          if (this.didBufferAllRequiredProps()) {
            this.pushOutBufferedProps();
          }
        }
      }
    }
    else if (this.propertyDelegate) {
      if (!this.didSync)
        this.pushOutBufferedProps();
      this.propertyDelegate.decodeSetPropertyMessage(buf);
    }
  }

  rxMIDI(buf) {
    if (! IsMOTU_Msg(buf.data))
      return;

    const msgType = GetMsgType(buf.data);
    if (msgType == undefined)
      return;
    
    switch(msgType) {
      case SysexRequestIDs.kSetProperty:
        this.rxPropertyChange(buf.data); 
        break;
      case SysexRequestIDs.kProtocolProbe:
        this.isMOTU = true;
        console.log("It's a MOTU device!");
        this.enableSysexPropertyAPI();

        break;
      default:
        console.log("unrecognized msg type: " + msgType);
        break;
    }
  }

  sendMIDI(buf) {
    this.output.send(buf);
  }

  probeSysexProtocol() {
    let buf = new Uint8Array(8);

    let i = 0;
    for (; i < kMOTUSysexHeaderBytes.length;++i) {
      buf[i] = kMOTUSysexHeaderBytes[i];
    }
    buf[i++] = ConstructRequestValue(kRequestDirectionHostToDevice, SysexRequestIDs.kProtocolProbe);
    buf[i++] = 0xF7;  // Sysex end

    this.sendMIDI(buf);
  }

  enableSysexPropertyAPI() {
    let buf = new Uint8Array(8);
    let i = 0;
    for (; i < kMOTUSysexHeaderBytes.length;++i) {
      buf[i] = kMOTUSysexHeaderBytes[i];
    }
    buf[i++] = ConstructRequestValue(kRequestDirectionHostToDevice, SysexRequestIDs.kEnableSysexAPI);
    buf[i++] = 0xF7;  // Sysex end

    this.sendMIDI(buf);
    this.didSync = false;
  }
};


class iOSMIDIProxy {
  constructor(jsHostMsgPipe) {
    this.rxMIDI = this.rxMIDI.bind(this);

    this.jsHostMsgPipe = jsHostMsgPipe;
  }

  rxMIDI(encodedBuf) {
    // This function is invoked by the iOS wrapper app.
    if (! this.propertyDelegate)
      return;

    // encodedBuf is a base64 encoded string.
    // Decode back to Uint8Array
    var decodedData = new Uint8Array(atob(encodedBuf).split('').map(function(char) {
        return char.charCodeAt(0);
    }));

    this.propertyDelegate.decodeSetPropertyMessage(decodedData);
  }

  sendMIDI(buf) {
    if (this.jsHostMsgPipe) {
      // Convert buf, which is a Uint8Array, to Base64-encoded string
      var base64String = btoa(String.fromCharCode.apply(null, buf));
      this.jsHostMsgPipe.postMessage(JSON.stringify({
        "midimessage":base64String
      }));
    }
  }

  enableSysexPropertyAPI() {
    ProxyLog("iosProxy - enableSysexPropertyAPI");
    if (this.jsHostMsgPipe) {
      this.jsHostMsgPipe.postMessage("EnableSysexAPI");
    }
  }
}


class WebMIDIManager {
  constructor(appdata) {
    this.enableWebMIDI = this.enableWebMIDI.bind(this);
    this.haveDevice = this.haveDevice.bind(this);
    this.onMIDISuccess = this.onMIDISuccess.bind(this);
    this.onMIDIFailure = this.onMIDIFailure.bind(this);
    this.onStateChange = this.onStateChange.bind(this);
    this.removeDevice = this.removeDevice.bind(this);
    this.enumerate = this.enumerate.bind(this);

    this.appdata = appdata;
    this.devices = [];
    this.motuDevices = [];
    this.enabled = false;
  }

  onMIDISuccess(midiAccess) {
    this.enumerate(midiAccess);
  }

  onMIDIFailure(msg) {
    console.error(`Failed to get MIDI access - ${msg}`);
  }

  onStateChange(e) {
    setTimeout(()=>{
      this.enumerate(this.midiAccess);
    }, 0);
  }

  haveDevice(input, output) {
    const idx = this.devices.findIndex((x)=>{
      return x.input == input && x.output == output;
    });
    return idx != -1;
  }

  removeDevice(input, output) {
    const isEqual = (x) => {
      return x.input == input && x.output == output;
    };

    {
      const idx = this.devices.findIndex(isEqual);
      if (idx != -1) {
        this.devices.splice(idx, 1);
      }
    }
    {
      const idx = this.motuDevices.findIndex(isEqual);
      if (idx != -1) {
        this.motuDevices.splice(idx, 1);
      }
    }
  }

  enumerate(midiAccess) {
    if (this.midiAccess == undefined) {
      this.midiAccess = midiAccess;
      this.midiAccess.onstatechange = this.onStateChange;
    }

    let devices = {};
    let wmds = [];

    for (const entry of midiAccess.inputs) {
      for (const port of entry) {
        devices[port.name] = {
          input: midiAccess.inputs.get(port.id)
        };
      }
    }

    for (const entry of midiAccess.outputs) {
      for (const port of entry) {
        if (devices.hasOwnProperty(port.name))
          devices[port.name].output = midiAccess.outputs.get(port.id);
        else
          devices[port.name] = {
            output: midiAccess.outputs.get(port.id)
          };
      }
    }

    const motuDeviceCallback = (wmd)=>{
      this.motuDevices.push(wmd);
      this.appdata.eventSys.dispatchASync("WebMIDIStateChanged");
    };

    let sendNotif = false;
    for (const d of Object.values(devices)) {
      if (d.input && d.output) {
        if (d.input.state == "connected" && d.output.state == "connected" && !this.haveDevice(d.input, d.output)) {
          this.devices.push(new WebMIDI_Device(d.input, d.output, motuDeviceCallback));
          sendNotif = true;
        }
        else if (d.input.state == "disconnected" || d.output.state == "disconnected" && this.haveDevice(d.input, d.output)) {
          this.removeDevice(d.input, d.output);
          sendNotif = true;
        }
      }
    }

    // Check for removed devices
    let removed = [];
    for (const d of this.devices) {
      const dlist = Object.values(devices);
      const idx = dlist.findIndex((x)=>{
        return d.input == x.input || d.output == x.output;
      });
      if (idx == -1) {
        removed.push(d);
        sendNotif = true;
      }
    }
    for (const r of removed) {
      const i1 = this.devices.indexOf(r);
      if (i1 != -1) {
        this.devices.splice(i1);
      }
      const i2 = this.motuDevices.indexOf(r);
      if (i2 != -1) {
        this.motuDevices.splice(i2);
      }
    }

    if (sendNotif) {
      this.appdata.eventSys.dispatchASync("WebMIDIStateChanged");
    }
  }

  enableWebMIDI() {
    if (this.enabled)
      return;

    if (PlatformSupportsWebMIDI()) {
      try {
        navigator.requestMIDIAccess({sysex: true}).then(this.onMIDISuccess, this.onMIDIFailure);
        this.enabled = true;
      }
      catch(e) {
        console.log("Failed to enable WebMIDI");
        console.log(e);
      }
    }
  }
}

class WeakPropertyDelegateWrapper {
  constructor (propMgr) {
    this.decodeSetPropertyMessage = this.decodeSetPropertyMessage.bind(this);
    this.mgr = new WeakRef(propMgr);
  }
  decodeSetPropertyMessage(x) {
    var mgr = this.mgr.deref();
    if (mgr) {
      mgr.decodeSetPropertyMessage(x);
    }
  }
}

// SysexPropertyMgr constructor can take a WebMIDI device
// or an iOSMIDIProxy.
class SysexPropertyMgr {
  constructor () {
    this.encodeDeviceMessage = this.encodeDeviceMessage.bind(this);
    this.decodeSetPropertyMessage = this.decodeSetPropertyMessage.bind(this);
    this.enableSysexPropIn = this.enableSysexPropIn.bind(this);
    this.disableSysexPropIn = this.disableSysexPropIn.bind(this);
  }

  setDevice(device) {
    this.device = device;

    if (device) {
      this.device.propertyDelegate = new WeakPropertyDelegateWrapper(this);
    }
  }

  encodeDeviceMessage(d) {
    let len = d.length - 4;
    let xtra = Math.ceil((4 + len) / 7);

    let array_length = 8 + 4 + len + xtra;
    let x = new Uint8Array(array_length);
    let n = 0;

    x[n++]  = 0xF0; // Start of sysex
    x[n++]  = 0x00; // MOTU
    x[n++]  = 0x00; // MOTU
    x[n++]  = 0x3B; // MOTU
    x[n++] = kDeviceSysexProtocolID[0];
    x[n++] = kDeviceSysexProtocolID[1];
    x[n++] = ConstructRequestValue(kRequestDirectionHostToDevice, SysexRequestIDs.kSetProperty);
    
    let shr = 7;
    let b7 = n++;

    // encode data
    for (let k = 0; k < d.length; ++k)
    {
        if (shr == 0)
        {
            b7 = n++;
            shr = 7;
        }

        const value = d[k];

        x[b7] |= (value & 0x80) >> shr--;
        x[n++] = value & 0x7F;
    }

    x[n++] = 0xF7; // End of sysex

    return x;
  }

  decodeSetPropertyMessage(data) {
    // ProxyLog("+decodeSetPropertyMessage");
    if (!this.propertyCallback)
    {
      // ProxyLog("No proprety callback")
      return;
    }

    const result = DecodeSetPropertySysexMessage(data);
    if (result) {
      this.propertyCallback(result);
    }
    this.propertyCallback(result);
  }

  enableSysexPropIn(cb) {
    ProxyLog("enableSysexPropIn");
    this.propertyCallback = cb;
    this.device.enableSysexPropertyAPI();
  }

  disableSysexPropIn() {
    ProxyLog("disableSysexPropIn");
    this.propertyCallback = undefined;
    // TODO: make an actual request to stop incoming sysex?
  }
}
