'use strict';

function getSampleRateIndex(sr) {
  var l = kSampleRates.length;
  for(var i = 0; i < l; ++i)
    if(kSampleRates[i].v == sr)
      return i;
  return 0;
}

function calcBusRudeSolo(device, v, b) {
  const device_props = device.properties;
  var o = device_props.get_mix_index(0, b);
  for(var c = 0; c < device_props.numInputs; ++c) {
    if(v[o + c] != 0)
      return true;
  }
  return false;
}

// private helper for adding host out input channels
function addBusHostInChannel(bus, iStereo, channels) {
  var chi = {};
  chi.chNum = channels.length;
  chi.stripType = "host";
  chi.type = bus.type;
  chi.bankChNum = bus.bankChNum;
  chi.stereoIndex = bus.istereoIndex;
  chi.isStereo = chi.stereoIndex == undefined ? 1 : iStereo[chi.stereoIndex];

  chi.pre = -1;
  chi.eq = -1;
  chi.gate = -1;
  chi.comp = -1;
  chi.name = bus.name + bus.bankChNum;
  chi.gainICh = bus.hoGainIndex; // channel row in fader
  chi.meters = [-1, -1, -1, -1, bus.hoMeter];

  if ((bus.bankChNum & 1) && !bus.isStereo && chi.isStereo) {
    // If the bus is not stereo, but is the "right channel" of a
    // stereo-linkable pair, and that host channel strip corresponding
    // to this output bus is stereo-linked, then don't add this
    // channel strip. It was already added when this function was
    // called for the "left channel", and we'll otherwise end up with
    // something like "Line 4-5".
    return;
  }

  channels.push(chi);


  // If the bus is not stereo-linked, do not add the "right" channel
  // because we'll end up with duplicated channel strips when
  // this function is called for the "right" channel of the output bus. 

  if(!chi.isStereo && bus.isStereo && !(bus.bankChNum & 1) ) {
    var chR = {};
    chR.chNum = channels.length;
    chR.stripType = "host";
    chR.type = bus.type;
    chR.bankChNum = bus.bankChNum + 1;
    chR.stereoIndex = bus.istereoIndex;
    chR.isStereo = false;

    chR.pre = -1;
    chR.eq = -1;
    chR.gate = -1;
    chR.comp = -1;
    chR.name = bus.name + bus.bankChNum + 1;
    chR.gainICh = bus.hoGainIndex + 1; // channel row in fader
    chR.meters = [-1, -1, -1, -1, bus.hoMeter + 1];

    channels.push(chR);
  }
}

class MixMgr {
  constructor(appdata) {
    this.meterStore = appdata.meterStore
    this.eventSys = appdata.eventSys;
    this.onDS = this.onDS.bind(this);
    this.dsSoloChanged = this.dsSoloChanged.bind(this);
    this.update = this.update.bind(this);

    this.setActiveBus = this.setActiveBus.bind(this);
    this.getActiveBus = this.getActiveBus.bind(this);
    this.getActiveBusInfo = this.getActiveBusInfo.bind(this);
    this.getNumBuses = this.getNumBuses.bind(this);
    this.getBusInfo = this.getBusInfo.bind(this);

    this.getBusRudeSolo = this.getBusRudeSolo.bind(this);
    this.clearBusSolos = this.clearBusSolos.bind(this);

    // active ch is for fxviews.
    this.incActiveCh = this.incActiveCh.bind(this);
    this.setActiveCh = this.setActiveCh.bind(this);
    this.getActiveCh = this.getActiveCh.bind(this);
    this.getNumCh = this.getNumCh.bind(this);
    this.getChInfo = this.getChInfo.bind(this);
    this.getChGainIndex = this.getChGainIndex.bind(this);

    this.rudeSolo = [];
    this.pending = false;
    this.channels = [];
    this.buses = [];

    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.datastore.addListener('kSampleRate', this.onDS);
    this.datastore.addListener('kiMixStereo', this.onDS);
    this.datastore.addListener('koMixStereo', this.onDS);
    this.datastore.addListener('kOpticalMode', this.onDS);
    this.datastore.addListener('kHostOutMixMode', this.onDS);
    this.datastore.addListener('kiMixSolo', this.dsSoloChanged);

    // activeBus: the index of the active bus info + the type and bankChNum
    this.activeBus = {i:0, type:"", bankChNum:0};
    // activeBus: the index of the active channel info + the type and bankChNum
    this.activeCh = {i:0, type:"", bankChNum:0};
    //setTimeout(this.update, 0); (called synchronously at the end of main)

    this.update();
  }
  incActiveCh(l) {
    var ac = l ? this.activeCh.i - 1 : this.activeCh.i + 1;
    if(ac >= 0 && ac < this.channels.length)
      this.setActiveCh(ac);
  }
  setActiveCh(ch) {
    if(isNaN(ch) || (ch >= this.channels.length))
      ch =0;
    if(this.activeCh.i != ch || this.activeCh.type != this.channels[ch].type || this.activeCh.bankChNum  != this.channels[ch].bankChNum) {
      this.activeCh.i = ch;
      this.activeCh.type = this.channels[ch].type;
      this.activeCh.bankChNum  = this.channels[ch].bankChNum;
      this.eventSys.dispatchASync('activeChChanged');
    }
  }
  getActiveCh() {
    return this.activeCh.i;
  }
  getNumCh() {
    return this.channels.length;
  }
  getChInfo(n) {
    return this.channels[n];
  }
  getChGainIndex(n, b) {
    // gainIndex is the only bus specific channel parameter. omit b and get active bus
    if(b == undefined)
        b = this.activeBus.i;
    return this.device.properties.get_mix_index(this.channels[n].gainICh, this.buses[b].gainOCh);
  }

  setActiveBus(b) {
    if(isNaN(b) || (b >= this.buses.length))
      b =0;
    if(this.activeBus.i != b || this.activeBus.type != this.buses[b].type || this.activeBus.bankChNum != this.buses[b].bankChNum) {
      this.setActiveCh(0);
      this.activeBus.i = b;
      this.activeBus.type = this.buses[b].type;
      this.activeBus.bankChNum  = this.buses[b].bankChNum;
      setTimeout(this.update, 0);
      this.eventSys.dispatchASync('activeBusChanged');
    }
  }
  getActiveBus() {
    return this.activeBus.i;
  }
  getActiveBusInfo() {
    return this.buses[this.activeBus.i];
  }
  getNumBuses() {
    return this.buses.length;
  }
  getBusInfo(n) {
    return this.buses[n];
  }

  onDS(k,v) {
    // I don't really care what changed here, just cue up a callback so we can
    // re layout the mixer. I do this asynchronously so if multiple things changed we
    // only relayout the ui once
    if(!this.pending) {
      this.pending = true;
      setTimeout(this.update, 0);
    }
    this.meterStore.meterInterval = 1/this.datastore.getKVArrayItem('kUIMeterFPS', 0);
  }
  update() {
    // update needs to be called when the bus or channel layout changes.
    // - at startup
    // - sample rate change
    // - when stereo changes
    // - when any bank (optical) changes

    this.pending = false;
    var sr = this.datastore.getKVArrayItem('kSampleRate', 0)
    var sri = getSampleRateIndex(sr);
    var is4XMult = getSampleRateMult(sr) == 4;
    var optMode = this.datastore.getKVArrayItem('kOpticalMode', 0);
    var iStereo = this.datastore.getKVArray('kiMixStereo');
    var oStereo = this.datastore.getKVArray('koMixStereo');
    var hostMixMode = this.datastore.getKVArrayItem('kHostOutMixMode', 0);

    this.channels = [];
    this.buses = [];

    // build a bus info for every stereo or mono bus
    const device_props = this.device.properties;
    var numob = device_props.busInfo.length;
    for (var i = 0; i < numob; i++) {
      var info = device_props.busInfo[i];
      var numCh = info.numCh;
      for(var bankChNum = 0; bankChNum < numCh; bankChNum++) {

        // read everything useful about the bus
        var bi = {};
        bi.type = info.type;
        bi.bankChNum = bankChNum;
        bi.stripType = 'bus';

        // if the bus support stereo/mono see if it's been set, otherwise it is always stereo
        if(info.stereo != undefined) {
          bi.stereoIndex = (info.stereo + bankChNum) & 0xfe; // always use the left ch
          bi.isStereo = oStereo[info.stereo + bankChNum];
        } else {
          bi.stereoIndex = undefined;
          bi.isStereo = 1;
        }

        // if the bus has a corresponding input from the host (main out from host) stereo/mono see if it's been set, otherwise it is always stereo
        if(info.istereo != undefined) {
          bi.istereoIndex = (info.istereo + bankChNum) & 0xfe; // always use the left ch
        } else {
          bi.istereoIndex = undefined;
        }

        // these are indexes into the ds arrays
        bi.eq = (info.busEq != undefined) ? (info.busEq + bankChNum) : -1;
        bi.name = (info.name != undefined) ? info.name + bankChNum : -1;
        bi.gainOCh = info.iBus + bankChNum; // offset to bus gain index (0-16)
        if (Array.isArray(info.meters[0])) {
          bi.meter = info.meters[0][bankChNum];
        } else {
          bi.meter = info.meters[0] + bankChNum;
        }
        bi.color = device_props.busColors[bi.gainOCh];
        bi.revRetGainICh = (bi.gainOCh * 32) + device_props.inputs.reverb.iGain;  // reverb inputs start at index 30
        if(info.iGain != undefined) {
          bi.hoGainIndex = info.iGain + bankChNum; // host out (input from host) gain index for this bus
          bi.hoMeter = info.meters[1] + bankChNum;
        }

        if(!is4XMult || (bi.type != device_props.MixTypes.Reverb))
          this.buses.push( bi );
        if(bi.isStereo)
          bankChNum++;
      }
    }

    // build a channel info for every stereo or mono channel
    const channel_counts = {};
    const numib = device_props.channelInfo.length;
    for (let i = 0; i < numib; i++) {
      const info = Object.assign(device_props.channelInfo[i]);
      const active = function() {
        if((info.type == device_props.MixTypes.ADATA) && (optMode === OpticalModes.TOSlink))
          return false;
        if((info.type == device_props.MixTypes.TOSlink) && (optMode === OpticalModes.ADAT))
          return false;
        return true;
      }();

      const ahi = this.buses[this.activeBus.i];
      if((info.type == device_props.MixTypes.Reverb) || (ahi && (ahi.type == device_props.MixTypes.Reverb) && (i == (numib - 1)))) {
        // Add host output channels before reverb channel, or if its the reverb mix after all other channels
        for (let b = 0; b < this.buses.length; ++b) {

          if ((this.buses[b].type == MixBusTypes.Phones) && is4XMult)
            continue;
          if (this.buses[b].type == MixBusTypes.Reverb)
            continue;

          if (b == 0) {
            addBusHostInChannel(this.buses[b], iStereo, this.channels);
          }
          else if (hostMixMode) {
            addBusHostInChannel(this.buses[b], iStereo, this.channels);
          }
          else {
            const bus = this.buses[b];
            const chIsStereo = iStereo[bus.istereoIndex];
            const isPartOfBusStereoPair = ahi.istereoIndex == bus.istereoIndex;
            if (chIsStereo && !ahi.isStereo) {
              if (bus.hoGainIndex == bus.istereoIndex && isPartOfBusStereoPair)
                addBusHostInChannel(this.buses[b], iStereo, this.channels);
            }
            else if (b == this.activeBus.i) {
              addBusHostInChannel(this.buses[b], iStereo, this.channels);
            }
          }
        }
      }

      if(active) {
        const baseIndex = function() {
          const testBase = channel_counts[info.type];
          if (testBase === undefined) {
            return 0;
          }
          else {
            return testBase;
          }
        }();
        const numCh = info.numCh[sri];
        for(var bankChNum = 0; bankChNum < numCh; bankChNum++) {
          // read everything useful about the channel
          let chi = {};
          chi.chNum = this.channels.length;
          chi.type = info.type;
          chi.bankChNum = baseIndex + bankChNum;
          // if the channel support stereo/mono see if it's been set, otherwise it is always stereo
          if(info.stereo != undefined) {
            chi.stereoIndex = (info.stereo + bankChNum) & 0xfe;
            chi.isStereo = iStereo[chi.stereoIndex];
          } else {
            chi.stereoIndex = undefined;
            chi.isStereo = 1;
          }
          // these are indexes into the ds arrays
          chi.pre = (info.pre != undefined) ? (info.pre + bankChNum) : -1;
          chi.eq = (info.eq != undefined) ? (info.eq + bankChNum) : -1;
          chi.gate = (info.gate != undefined) ? (info.gate + bankChNum) : -1;
          chi.comp = (info.comp != undefined) ? (info.comp + bankChNum) : -1;
          chi.name = info.name + bankChNum;
          chi.gainICh = (info.iGain + bankChNum); // channel row in fader
          chi.meters = info.meters.slice(); //[...info.meters];
          chi.meters[0] += (chi.meters[0] != -1) ? bankChNum : 0;
          chi.meters[1] += (chi.meters[1] != -1) ? bankChNum : 0;
          chi.meters[2] += (chi.meters[2] != -1) ? bankChNum : 0;
          chi.meters[3] += (chi.meters[3] != -1) ? bankChNum : 0;
          if (Array.isArray(chi.meters[4])) {
            chi.meters[4] = chi.meters[4][bankChNum];
          } else {
            chi.meters[4] += (chi.meters[4] != -1) ? bankChNum : 0;
          }
          chi.stripType = (chi.type != device_props.MixTypes.Reverb) ? "input" : "reverb";

          if(chi.type  != device_props.MixTypes.Reverb)
            this.channels.push(chi);
          else if((this.activeBus.type != MixBusTypes.Reverb) && !is4XMult)
            this.channels.push(chi);

          if(chi.isStereo)
            bankChNum++;
        }
        channel_counts[info.type] = baseIndex + numCh;
      }
    }

    // deal with the possiblity that the active bus may have disappeared (stereo)
    const ab = this.activeBus.i;
    if(this.buses.length <= ab)
      this.setActiveBus(this.buses.length - 1);
    else if((this.buses[ab].type != this.activeBus.type) || (this.buses[ab].bankChNum != this.activeBus.bankChNum)) {
      this.setActiveBus(ab > 0 ? ab-1 : 0);
    }

    // deal with the possiblity that the channel may have disappeared (stereo)
    const ac = this.activeCh.i;
    if(this.channels.length <= ac)
      this.setActiveCh(this.channels.length - 1) ;
    else if((this.channels[ac].type != this.activeCh.type) || (this.channels[ac].bankChNum != this.activeCh.bankChNum)) {
      this.setActiveCh(ac > 0 ? ac-1 : 0);
    }

    this.meterStore.clear();
    this.eventSys.dispatchASync('mixCfgChanged');
    this.eventSys.dispatchASync('activeChChanged');
  }
  dsSoloChanged(k, v) {
    // cache if the bus has any solo'd channels
    for(let b = 0; b < 16; ++b)
      this.rudeSolo[b] = calcBusRudeSolo(this.device, v,b);
    this.eventSys.dispatchASync('soloChanged');
  }
  getBusRudeSolo(b) {
    return this.rudeSolo[b];
  }
  clearBusSolos(b) {
    const device_props = this.device.properties;
    const o = device_props.get_mix_index(0, b);
    for(let c = 0; c < device_props.numInputs; ++c)
      this.datastore.setV('kiMixSolo', o + c, 0);
  }
}


// helper to get channel/bus name
function getChannelName(datastore, dsKey, info) {
  var n = "";
  if ((info.type != "Reverb") && (info.stripType != "host")) {
    n = datastore.getKVArrayItem(dsKey, info.name);
  }
  if((n == undefined) || n == "")
  {
    // the default name comes from the info.type
    n = info.type;
    var lch = info.bankChNum + 1;
    if(info.type == "Line")
      lch += 2;
    else if (info.type == "Phones")
      lch = ((lch - 1) / 2) + 1;

    if(info.type != "Reverb") {
      if(info.isStereo) {
        if (info.type == "Phones") {
          if (datastore.getKVArrayItem("kModelID", 0).startsWith("828"))
            n += " " + lch;
        }
        else
          n += " " + lch + "-" + (lch+1);
      }
      else {
        // Mono
        n += " " + lch;
      }
    }
  }

  return n;
}

