'use strict';

function send(url) {
  return new Promise(function(resolve, reject){
    var xhr = new XMLHttpRequest();
    function ensureReadiness() {
    if(xhr.readyState < 4)
      return;
    if((xhr.status !== 200) && (xhr.status !== 304))
      reject();
    if(xhr.readyState === 4)
      resolve(xhr.response);
    }
    xhr.onreadystatechange = ensureReadiness;
    xhr.open('GET', url, true);
    xhr.timeout = 5000;
    xhr.send('');
  });
}

class NewFirmwarePoller
{
  constructor(callback) {
    this.getLatest = this.getLatest.bind(this);
    this.families = [
      { models: ['ultralite-mk5l'], latest: 0 },
      { models: ['828'], latest:0},
      { models: ['gen5driver-mac'], latest: 0 },
      { models: ['gen5driver-win'], latest: 0 },
    ];
    var funcs = [];
    for (var i in this.families) {
      var lower = encodeURIComponent(this.families[i].models[0].toLowerCase());
      var url = 'http://firmware.motu.com/api/device/' + lower + '/firmware';
      ((url, family) => {
        funcs.push(()=>send(url, 0));
        funcs.push((r)=>{var j = JSON.parse(r); if (j.length) family.latest = j[0]});
      })(url, this.families[i]);
    }
    funcs.push(()=>callback());
    funcs.reduce((prev, cur) => prev.then(cur), Promise.resolve());
  }
  getLatest(model) {
    for (var i in this.families) {
      if (this.families[i].models.indexOf(model) > -1)
        return this.families[i].latest;
    }
  }
}

class Updater {
  constructor(appdata) {
    this.sendMsg = this.sendMsg.bind(this);
    this.isAvailable = this.isAvailable.bind(this);
    this.start = this.start.bind(this);
    this.refresh = this.refresh.bind(this);
    this.updateLatestAvailable = this.updateLatestAvailable.bind(this);

    this.appdata = appdata;
    this.services = {};
    this.images = {};
    this.started = false;

    this.latestPoller = new NewFirmwarePoller(this.updateLatestAvailable);
    this.latestFirmware = new Map();
    this.latestDriver = "";
    this.driverVersion = "";

    this.hasFile = false;
    this.serial = this.versionFW = this.fileVersionFW = this.versionFPGA = this.fileVersionFPGA = "";
    this.updating = false;
  }
  updateLatestAvailable() {
    var refresh = false;
    var latest = this.latestPoller.getLatest('ultralite-mk5l');
    if (latest && latest.version) {
      this.latestFirmware.set(DeviceTypes.UltraLite, latest.version);
      refresh = true;
    }
    latest = this.latestPoller.getLatest('828');
    if (latest && latest.version) {
      this.latestFirmware.set(DeviceTypes.k828, latest.version);
      refresh = true;
    }
    if (/(Mac)/i.test(navigator.platform)) {
      latest = this.latestPoller.getLatest('gen5driver-mac');
      if (latest && latest.version) {
        this.latestDriver = latest.version;
        refresh = true;
      }
    }
    else {
      latest = this.latestPoller.getLatest('gen5driver-win');
      if (latest && latest.version) {
        this.latestDriver = latest.version;
        refresh = true;
      }
    }
    if (refresh)
      this.refresh();
  }
  updateDriverVersion(version) {
    this.appdata.eventSys.dispatchASync('driverVersion', version);
    this.driverVersion = version;
    if (!(/(Mac)/i.test(navigator.platform))) {
      if ((version.match(/\./g) || []).length == 3) {
        var parts = version.split('.');
        this.driverVersion = parts[0] + '.' + parts[1] + '.0+' + String(Number(parts[2]) * 10000 + Number(parts[3]));
      }
    }
    this.refresh();
  }
  constructSerialNumber() {
    if (this.modelID.search("828") == -1) {
      // It's an Ultralite-mk5
      return (this.appdata.datastore.getKVArrayItem('kModelID', 0) + this.mac[3].toString(16) + this.mac[4].toString(16) + this.mac[5].toString(16)).toUpperCase();
    }
    // It's an 828
    let sn = this.appdata.datastore.getKVArrayItem('kDevUID', 0);
    if (sn.length == 0 || sn[0] == 0) {
      return this.appdata.datastore.getKVArrayItem('kModelID', 0).toUpperCase() + "FFFFFF"
    }
    let sn_str = "";
    for (const i of sn) {
      sn_str += String.fromCharCode(i);
    }
    return sn_str.toUpperCase();
  }
  refresh() {
    if (this.appdata.device.isConnected()) {
      this.modelName = this.appdata.datastore.getKVArrayItem('kDevName', 0);
      this.modelID = this.appdata.datastore.getKVArrayItem('kModelID', 0);
      this.versionFW = this.appdata.datastore.getKVArrayItem('kDevVersion', 0).toUpperCase();
      this.versionFPGA = this.appdata.datastore.getKVArrayItem('kDevVersion', 1).toUpperCase();

      // Serial number is "ULM5", "UL5E" or "UL5L" + lower 3 bytes, UID is MAC with fffe in center
      this.mac = this.appdata.datastore.getKVArrayItem('kDeviceMAC', 0);
      this.serial = this.constructSerialNumber();

      this.uid =
        this.mac[0].toString(16).padStart(2, '0') +
        this.mac[1].toString(16).padStart(2, '0') +
        this.mac[2].toString(16).padStart(2, '0') + "fffe" +
        this.mac[3].toString(16).padStart(2, '0') +
        this.mac[4].toString(16).padStart(2, '0') +
        this.mac[5].toString(16).padStart(2, '0');
    }
    this.appdata.eventSys.dispatchASync('refreshUpdateClients');
  }
  start() {
    if(this.appdata.device.hasHostMsgPipe() && !this.appdata.device.has_iOS_hostMsgPipe()) {
      this.appdata.device.sendHostMsgPipeMsg("start");
      this.appdata.eventSys.addListener('updateChanged', this.refresh);
      this.started = true;
    }
    this.appdata.datastore.addListener('kDeviceMAC', this.refresh);
    this.appdata.datastore.addListener('kDevUID', this.refresh);
    this.appdata.datastore.addListener('kDevName', this.refresh);
    this.appdata.datastore.addListener('kDevVersion', this.refresh);
  }

  isAvailable(){
    return this.started;
  }
  openImageFromDisk() {
    this.sendMsg("open " + this.modelID );
  }
  openDriverDownload() {
    this.sendMsg("driver")
  }

  // Sends JSON to updater running in objectiveC/C++ or electron
  sendMsg(msg) {
    // messages:
    // "start" : start search for updatable devices
    // "get" : get the service list
    // "open" : run the open file dialog and verify firmwware image
    // "version&modelName" : get the version string for the firmware for specifed model
    // "update&firmware&modelName&uid" : Firmware update model with UID
    // "update&fpga&modelName&uid" : FPGA update model with UID
    if(this.started)
      this.appdata.device.sendHostMsgPipeMsg(msg);
  }
}


class UpdateView extends View {
  constructor(l, t, appdata, m) {
    m = m ? m : 0;
    super(l, t, 370, 180 + m);
    this.margin = m;
    this.appdata = appdata;

    this.isLocalDevice = this.isLocalDevice.bind(this);

    // called by the updater
    appdata.eventSys.addListener('refreshUpdateClients', this.requestRender);

    this.setBG(kEmptyBG);

    this.wantsMouse = true;
    this.chooseRect = new Rect(this.l + this.margin + 160, this.t + 100, 70, 30);
    this.driverRect = new Rect(this.l + this.margin + 160, this.t + 150, 70, 30);

    this.enableClipping(true);
  }

  offset(dx,dy) {
    if (this.chooseRect)
      this.chooseRect.offset(dx,dy);
    if (this.driverRect)
      this.driverRect.offset(dx,dy);
    super.offset(dx,dy);
  }

  handleMouseDown(xy) {
    if(this.appdata.updater.isAvailable() && this.appdata.device.isConnected()) {
      if(this.chooseRect.contains(xy.x, xy.y)) {
        this.appdata.updater.openImageFromDisk();
      }
      if(this.driverRect.contains(xy.x, xy.y)) {
        this.appdata.updater.openDriverDownload();
      }
    }
  }

  constructVersionString() {
    let verStr = "";
    const modelID = this.appdata.datastore.getKVArrayItem('kModelID', 0);
    if (modelID.search("828") == -1) {
      verStr += (this.appdata.updater.versionFW + ":" + this.appdata.updater.versionFPGA).toUpperCase();
    }
    else {
      verStr = this.appdata.updater.versionFW;
    }
    return verStr;
  }

  isLocalDevice() {
    const dev = this.appdata.device.netDevice;
    if (dev) {
      return dev.ip && (dev.ip.startsWith("127.0.0.1") || dev.ip.toLowerCase().startsWith("localhost"));
    }
    return false;
  }

  render(c) {
    if(!this.appdata.device.isConnected())
      return;

    c.fillStyle = "white";
    c.font = "300 15px Arial";
    c.textAlign = 'left';
    c.fillText("Serial Number:", this.l + this.margin, this.t + 20);
    c.fillText("Firmware Version:", this.l + this.margin, this.t + 70);

    if (this.appdata.device.hasHostMsgPipe() && !this.appdata.device.has_iOS_hostMsgPipe() && this.isLocalDevice()) {
      c.fillText("Firmware Updater:", this.l + this.margin, this.t + 120);
      c.fillText("CueMix Download:", this.l + this.margin, this.t + 170);
    }
    c.fillStyle = kToggleDefault;
    c.fillText(this.appdata.updater.serial.toUpperCase(), this.l + this.margin + 160, this.t + 20);
    var updateAvailable = false;

    let verStr = this.constructVersionString();

    if(verStr.length == 0 || verStr == ":") {
        verStr = "Needs Update";
        updateAvailable = true;
    }
    else if (this.appdata.updater.latestFirmware.has(this.appdata.device.type) && (this.appdata.updater.latestFirmware.get(this.appdata.device.type).toUpperCase() != this.appdata.updater.versionFW.toUpperCase()))
        updateAvailable = true;

    if(this.appdata.updater.hasFile) {
      if(this.appdata.updater.versionFW == this.appdata.updater.fileVersionFW)
        verStr += " up to date";
      else
        verStr += " updatable";
    }
    c.fillText(verStr, this.l + this.margin  + 160, this.t + 70);

    if(this.appdata.updater.isAvailable() && this.isLocalDevice()) {
      drawButton(c, this.chooseRect, "Open...", false, "black");
      drawButton(c, this.driverRect, "Open...", false, "black");
      c.fillStyle = "RGB(20,122,255)";
      c.textAlign = 'center';
      if (updateAvailable) {
        c.fillText("Update Available!", this.l + this.margin + 300, this.t + 120);
      }
      if (this.appdata.updater.latestDriver && this.appdata.updater.driverVersion && (this.appdata.updater.latestDriver != this.appdata.updater.driverVersion)) {
        c.fillText("Update Available!", this.l + this.margin + 300, this.t + 170);
      }
    }
  }
}
