'use strict';

////////// helper functions //////////////

function getSampleRateMult(sr) {
   return (sr <= 48000) ? 1 : ((sr <= 96000) ? 2 : 4);
}

// convert float to/from 8.24 int
const k824_1 = 0x01000000;
function To824(v) {
  return Math.round(v * k824_1);
}
function From824(v) {
  return v / k824_1;
}

function From_dB(v) {
  return Math.pow(10, 0.05 * v);
};
const k20OverLog10 = 20 / Math.log(10);
function To_dB(v) {
  return k20OverLog10 * Math.log(v);
};

// convert a float 0-1 to 16bits 0-ffff
function To16(v) {
  return Math.round(k16Bit * v) & k16Bit;
}
// convert a 16bit 0-ffff to a float range 0-1
const k16Bit = 0x0000ffff;
function From16(v) {
  return (v & k16Bit) / k16Bit;
}

function CheckArg(arg, argType) {
    if(!(arg instanceof argType))
      throw 'Bad Arg expected ' + argType.name;
}

function trimFloat(v, nd) {
  if(!isNaN(v))
    v = v.toFixed(nd);
  return v;
}

// updtime
var perf = window.performance;
function perfSeconds() { return perf.now()/1000.0; }

/* View in fullscreen */
function enterFullscreen() {
  var elem = document.documentElement;
  if (elem.requestFullscreen) {
    elem.requestFullscreen();
  } else if (elem.mozRequestFullScreen) { /* Firefox */
    elem.mozRequestFullScreen();
  } else if (elem.webkitRequestFullscreen) { /* Chrome, Safari and Opera */
    elem.webkitRequestFullscreen();
  } else if (elem.msRequestFullscreen) { /* IE/Edge */
    elem.msRequestFullscreen();
  }
}

/* Close fullscreen */
function exitFullscreen() {
  var elem = document.documentElement;
  if (document.exitFullscreen) {
    document.exitFullscreen();
  } else if (document.mozCancelFullScreen) { /* Firefox */
    document.mozCancelFullScreen();
  } else if (document.webkitExitFullscreen) { /* Chrome, Safari and Opera */
    document.webkitExitFullscreen();
  } else if (document.msExitFullscreen) { /* IE/Edge */
    document.msExitFullscreen();
  }
}

function getOutputPairChName(datastore, default_names, name_index) {
  var name = datastore.getKVArrayItem('koChannelNames', name_index);
  if (datastore.getKVArrayItem('koMixStereo', name_index)) {
    return (name ? name : default_names.stereo) + " Mix (L-R)";
  }
  else {
    var other = datastore.getKVArrayItem('koChannelNames', name_index+1);
    other = (other ? other : default_names.mono_r) + " Mix (R)";
    return (name ? name : default_names.mono_l) + " Mix (L) - " + other;
  }
}

const getMainOutputPairName = function() {
  const defaults = {
    stereo: "Main 1-2",
    mono_l: "Main 1",
    mono_r: "Main 2"
  };
  return function(device, datastore) {
    return getOutputPairChName(datastore, defaults, device.properties.nameIndices.main.output);
  };
}();

const getLine3OutputPairName = function() {
  const defaults = {
    stereo: "Line 3-4",
    mono_l: "Line 3",
    mono_r: "Line 4"
  };
  return function(device, datastore) {
    return getOutputPairChName(datastore, defaults, device.properties.nameIndices.line.output);
  };
}();

const getLine5OutputPairName = function() {
  const defaults = {
    stereo: "Line 5-6",
    mono_l: "Line 5",
    mono_r: "Line 6"
  };
  return function(device, datastore) {
    return getOutputPairChName(datastore, defaults, device.properties.nameIndices.line.output+2);
  };
}();

const getLine7OutputPairName = function() {
  const defaults = {
    stereo: "Line 7-8",
    mono_l: "Line 7",
    mono_r: "Line 8"
  };
  return function(device, datastore) {
    return getOutputPairChName(datastore, defaults, device.properties.nameIndices.line.output+4);
  };
}();

const getLine9OutputPairName = function() {
  const defaults = {
    stereo: "Line 9-10",
    mono_l: "Line 9",
    mono_r: "Line 10"
  };
  return function(device, datastore) {
    return getOutputPairChName(datastore, defaults, device.properties.nameIndices.line.output+6);
  };
}();


function getPhonesOutputPairName(device, datastore, num) {
  const num_phones = device.properties.outputs.phones.numCh / 2;

  const base = num * 2;
  const friendlyNum = num_phones > 1 ? String(num + 1) : "";

  const defaults = {
    stereo: "Phones " + friendlyNum,
    mono_l: "Phones " + friendlyNum + " L",
    mono_r: "Phones " + friendlyNum + " R"
  };
  return getOutputPairChName(datastore, defaults, device.properties.nameIndices.phones.output + base);
}



const getPhones1OutputPairName = function(device, datastore) {
  return getPhonesOutputPairName(device, datastore, 0);
}

////////// utility classes //////////////
class AnimationFrameDispatcher {
  constructor() {
    this.queue = this.queue.bind(this);
    this.queueMeters = this.queueMeters.bind(this);
    this.dispatch = this.dispatch.bind(this);
    this.metersCB = null;
    this.callbacks = [];
    this.updateTS = 0;
    window.requestAnimationFrame(this.dispatch);
  }
  queueMeters(cb) {
    this.metersCB = cb;
  }
  queue(cb) {
    if(this.callbacks.indexOf(cb) == -1) {
      this.callbacks.push(cb);
    }
  }
  dispatch(timestamp) {
    timestamp /= 1000.0
    if(this.updateTS <= timestamp) {
      // do meters first so they can be rendered this pass
      if(this.metersCB != null)
        this.metersCB(timestamp);

      var c = this.callbacks;
      this.callbacks = [];
      c.forEach(function(cb) { cb(timestamp) });

      this.updateTS = timestamp + kMaxRefreshInterval;
    }
    window.requestAnimationFrame(this.dispatch);
  }
}


class Events {
  constructor() {
    this.addListener = this.addListener.bind(this);
    this.removeListener = this.removeListener.bind(this);
    this.dispatchSync = this.dispatchSync.bind(this);
    this.dispatchASync = this.dispatchASync.bind(this);
    this.dispatchPending = this.dispatchPending.bind(this);

    this.events = {}; // properties will be the event key holding an array of callbacks for the event
    this.pending = {};
    this.timerQueued = false;
  }
  // FWIW, the callback will be called with the key and the value.
  addListener(ev, l) {
    if(l == undefined)
      throw 'bad callback';

    if(!this.events.hasOwnProperty(ev))
      this.events[ev] = [];
    var o = this.events[ev];
    if(o.indexOf(l) == -1)
      o.push(l);
  }
  removeListener(ev, l) {
    var o = this.events[ev];
    if(o != undefined) {
      var i = o.indexOf(l);
      if(i > -1)
        o.splice(i, 1);

      if(o.length == 0)
        delete this.events.ev;
    }
  }
  dispatchSync(ev, v, setByObject) {
    var o = this.events[ev];
    if(o != undefined) {
      var n = o.length;
      for(var i = 0; i < n; i++)
        o[i](ev, v, setByObject);
    }
  }
  dispatchASync(ev, v, setByObject) {
    this.pending[ev] = [v, setByObject];
    if(!this.timerQueued) {
      this.timerQueued = true;
      setTimeout(this.dispatchPending, 0);
    }
  }
  dispatchPending() {
    var p = this.pending;
    this.timerQueued = false;
    this.pending = {};
    for (let [ev, v] of Object.entries(p))
      this.dispatchSync(ev, v[0], v[1]);
  }
}

class ValueProvider {
  constructor(val) {
    this.value = val;

    this.set = this.set.bind(this);
    this.broadcast = this.broadcast.bind(this);
    this.addListener = this.addListener.bind(this);
    this.removeListener = this.removeListener.bind(this);

    this.listeners = new Set();
    this.isBroadcasting = false;
  }

  addListener(obj) {
    this.listeners.add(obj);
  }
  removeListener(obj) {
    this.listeners.delete(obj);
  }
  set(val) {
    if (this.value != val) {
      this.value = val;
      this.broadcast();
    }
  }
  broadcast() {
    const listeners = this.listeners.entries();
    for (const [l, _] of listeners) {
      if (this.listeners.has(l)) { // If the listener wasn't removed since function entry
        l(this.value);
      }
    }
  }
}

class Rect {
  constructor(l, t, w, h) {
    this.uRBCXCY = this.uRBCXCY.bind(this);
    this.setR = this.setR.bind(this);
    this.setL = this.setL.bind(this);
    this.setT = this.setT.bind(this);
    this.setLT = this.setLT.bind(this);
    this.offset = this.offset.bind(this);
    this.setW = this.setW.bind(this);
    this.setH = this.setH.bind(this);
    this.setWH = this.setWH.bind(this);
    this.setLTWH = this.setLTWH.bind(this);
    this.contains = this.contains.bind(this);
    this.intersects = this.intersects.bind(this);
    this.centerOn = this.centerOn.bind(this);
    this.centerOnX = this.centerOnX.bind(this);
    this.centerOnY = this.centerOnY.bind(this);

    this.l = (l == undefined) ? 0 : l;
    this.t = (t == undefined) ? 0 : t;
    this.w = (w == undefined) ? 0 : w;
    this.h = (h == undefined) ? 0 : h;
    this.r = this.l + this.w;
    this.b = this.t + this.h;
    this.cx = this.l + this.w/2;
    this.cy = this.t + this.h/2;
  }
  uRBCXCY() {
    this.r = this.l + this.w;
    this.b = this.t + this.h;
    this.cx = this.l + this.w/2;
    this.cy = this.t + this.h/2;
  }
  copyR() {
    return new Rect(this.l, this.t, this.w, this.h);
  }
  setR(r) {
    this.setLTWH(r.l, r.t, r.w, r.h);
  }
  setL(l)
  {
    this.l = l;
    this.uRBCXCY();
  }
  setT(t)
  {
    this.t = t;
    this.uRBCXCY();
  }
  setLT(l,t) {
    this.l = l;
    this.t = t;
    this.uRBCXCY();
  }
  offset(dx,dy) {
    this.l = this.l + dx;
    this.t = this.t + dy;
    this.uRBCXCY();
  }
  setW(w) {
    this.w = w;
    this.uRBCXCY();
  }
  setH(h) {
    this.h = h;
    this.uRBCXCY();
  }
  setWH(w,h) {
    this.w = w;
    this.h = h;
    this.uRBCXCY();
  }
  setLTWH(l,t,w,h) {
    this.l = l;
    this.t = t;
    this.w = w;
    this.h = h;
    this.uRBCXCY();
  }
  contains(x, y) {
    return (this.l <= x) && (this.r > x) && (this.t <= y) && (this.b > y);
  }
  containsR(r) {
    return (this.l <= r.l) && (this.r >= r.r) && (this.t <= r.t) && (this.b >= r.b);
  }
  intersects(r) {
    return ! ((this.r < r.l) || (this.l > r.r) || (this.b < r.t) || (this.t > r.b));
  }
  centerOn(x, y) {
    const dx = this.cx - x;
    const dy = this.cy - y;
    this.offset(-dx, -dy);
  }
  centerOnX(x) {
    const dx = this.cx - x;
    this.offset(-dx, 0);
  }
  centerOnY(y) {
    const dy = this.cy - y;
    this.offset(0, -dy);
  }
}

class List {
  constructor() {
    this.items = [];
    this.clear = this.clear.bind(this);
    this.add = this.add.bind(this);
    this.get = this.get.bind(this);
    this.length = this.length.bind(this);
  }
  clear() { this.items = []; }
  add(v) { if(this.items.indexOf(v) == -1) { this.items.push(v); } }
  remove(v) {
    var i = this.items.indexOf(v);
    if(i != -1)
      this.items.splice(i,1);
  }
  get(i) { return this.items[i]; }
  length() { return this.items.length; }
}



class KeyDownFilter extends List {
  constructor() {
    super();
    this.keyDown = this.keyDown.bind(this);
    document.addEventListener('keydown', this.keyDown);
  }

  keyDown(e) {
    e = e || window.event;

    var keyCode = undefined;
    if (e.keyCode == 13) // escape
        keyCode = "enter";
    if (e.keyCode == 27) // escape
        keyCode = "escape";
    if (e.keyCode == 37) // larrow
        keyCode = "larrow";
    if (e.keyCode == 39) // rarrow
        keyCode = "rarrow";

    if (keyCode != undefined) {
      // go through the list backwards to find the last handler
      var n = this.length();
      for(var i = n-1; i >= 0; i--) {
        if(this.get(i).filterKey(keyCode))
          return;
      }
    }
  }
}

function getExp(cur, exp) {
  if(exp == "toggle")
    return (cur == "on") ? "off" : "on";
  if(exp == "on")
    return "on"
  return "off";
}

function stringDimensions(str, c, root)
{
  const stringHeight = (metrics)=>{
    return (metrics.actualBoundingBoxAscent + metrics.actualBoundingBoxDescent) / root.divScale;
  };
  const stringWidth = (metrics)=>{
    return metrics.width / root.divScale;
  };

  const fontHeight = (metrics) =>{
    return (metrics.fontBoundingBoxAscent + metrics.fontBoundingBoxDescent) / root.divScale;
  }
  const metrics = c.measureText(str);

  return {
    width: stringWidth(metrics),
    height: stringHeight(metrics)
  };
}