'use strict';


class DragTask {
  constructor(xy, cb) {
    this.updateDrag = this.updateDrag.bind(this);
    this.getDelta = this.getDelta.bind(this);

    this.startPos = xy;
    this.lastPos = xy;
    this.isDragging = false;
    this.cb = cb; // callback
  }

  updateDrag(xy) {
    this.lastPos = xy;

    if (!this.isDragging) {
      let dx = xy.x - this.startPos.x;
      let dy = xy.y - this.startPos.y;
      let mag = Math.sqrt(dx*dx + dy*dy);
      const kThreshold = 5;
      if (mag > kThreshold) {
        this.isDragging = true;
      }
    }

    if (this.isDragging) {
      try {
        this.cb(xy);
      }
      catch(e) {
        //...
      }
    }
  }

  getDelta() {
    return {
      x : this.lastPos.x - this.startPos.x,
      y : this.lastPos.y - this.startPos.y
    };
  }
};

const Swipe = {
  Left: 0,
  Right: 1,
  Up: 2,
  Down: 3
};


class SwipeDetector extends DragTask{
  constructor(xy) {
    super(xy, (p)=>{ this.touchMoved(p); });
    this.touchMoved = this.touchMoved.bind(this);
    this.touchEnded = this.touchEnded.bind(this);
    this.deltaToDirection = this.deltaToDirection.bind(this);
    this.horizontal = this.horizontal.bind(this);
    this.getAxisDelta = this.getAxisDelta.bind(this);
    this.getSwipe = this.getSwipe.bind(this);

    this.lastTime = Date.now();
    // this.lastPos = xy;
    this.lastSnapshotPos = xy;
    this.velocity = undefined; // will be px per sec
    this.direction = undefined;
  }

  deltaToDirection(xy) {
    let dx = xy.x;
    let dy = xy.y;

    if (dx == 0 && dy == 0)
      return undefined;

    let horiz = Math.abs(dx) > Math.abs(dy);
    if (horiz) {
      if (dx > 0)
        return Swipe.Right;
      return Swipe.Left;
    }
    if (dy > 0)
      return Swipe.Down;
    return Swipe.Up;
  }

  horizontal(swipeDirection) {
    return swipeDirection == Swipe.Left || swipeDirection == Swipe.Right;
  }

  getAxisDelta(delta) {
    let horiz = this.horizontal(this.deltaToDirection(delta));
    if (horiz)
      return delta.x;
    return delta.y;
  }

  touchMoved(xy) {
    let dx = xy.x - this.lastSnapshotPos.x;
    let dy = xy.y - this.lastSnapshotPos.y;

    let now = Date.now();

    let delta = {
      x: dx,
      y: dy
    };

    this.lastSnapshotPos = xy;

    let dir = this.deltaToDirection(delta);
    if (dir != undefined)
      this.direction = dir;

    let axisDelta = this.getAxisDelta(delta);
    this.velocity = (axisDelta * 1000.0) / (now - this.lastTime);

    this.lastTime = now;
  }

  touchEnded(xy) {
    let now = Date.now();
    if (now - this.lastTime <= 1000.0/30.0) {
      return;
    }
    this.updateDrag(xy);
  }

  getSwipe() {
    const kVelocityThreshold = 100;

    console.log("getSwipe: direction: " + String(this.direction) + ", velocity: " + String(this.velocity));

    if (this.direction != undefined && Math.abs(this.velocity) > kVelocityThreshold) {
      return this.direction;
    }
    return undefined;
  }

};


class View extends Rect{
  constructor(l,t,w,h) {

    super(l, t, w, h);

    this.setRoot = this.setRoot.bind(this);
    this.addSub = this.addSub.bind(this);
    this.removeSub = this.removeSub.bind(this);
    this.setVis = this.setVis.bind(this);
    this.visChanged = this.visChanged.bind(this);
    this.offset = this.offset.bind(this);
    this.render = this.render.bind(this);
    this.renderSelfAndSubs = this.renderSelfAndSubs.bind(this);
    this.requestRender = this.requestRender.bind(this);
    this.getMouseHandler = this.getMouseHandler.bind(this);
    this.handleDblClick = this.handleDblClick.bind(this);
    this.handleMouseDown = this.handleMouseDown.bind(this);
    this.handleMouseMoved = this.handleMouseMoved.bind(this);
    this.handleMouseUp = this.handleMouseUp.bind(this);
    this.findTop = this.findTop.bind(this);
    this.setBG = this.setBG.bind(this);
    this.dbgStrokeRect = this.dbgStrokeRect.bind(this);
    this.enableClipping = this.enableClipping.bind(this);
    this.setAlpha = this.setAlpha.bind(this);
    this.setEnabled = this.setEnabled.bind(this);
    this.getEffectiveAlpha = this.getEffectiveAlpha.bind(this);
    this.getCumulativeEffectiveAlpha = this.getCumulativeEffectiveAlpha.bind(this);
    this.cumulativeEffectiveAlphaChanged = this.cumulativeEffectiveAlphaChanged.bind(this);
    this.getNearestOpaqueParent = this.getNearestOpaqueParent.bind(this);

    // note  that coordinates are always relative to the root view
    this.root = undefined;
    this.subViews = new List;
    this.wantsMouse = false;
    this.isVis = true;
    this.isPVis = true;
    this.isClippingEnabled = false;
    this.alpha = 1.0;
    this.enabled = true;
  }

  setRoot(root) {
    this.root = root;
    var n = this.subViews.length();
    for(var i = 0; i < n; i++)
      this.subViews.get(i).setRoot(root);
  }
  addSub(v) {
    this.subViews.add(v);
    v.setRoot(this.root);
    v.setParentVis(this.isVis && this.isPVis);
    v.parentView = this;
  }
  removeSub(v) {
    this.subViews.remove(v);
    v.setRoot(undefined);
    v.setParentVis(false);
    v.parentView = undefined;
    this.requestRender();
  }
  visChanged(){
    var n = this.subViews.length();
    for(var i = 0; i < n; i++)
      this.subViews.get(i).setParentVis(this.isVis && this.isPVis);

    if(this.isVis && this.isPVis)
      this.requestRender();
  }
  setParentVis(isPVis) {
    if(this.isPVis != isPVis) {
      this.isPVis = isPVis;
      this.visChanged();
    }
  }
  setVis(isVis) {
    if(this.isVis != isVis) {
      this.isVis = isVis;
      this.visChanged();
    }
  }
  setEnabled(e) {
    const wasEnabled = this.enabled;
    this.enabled = !!e;
    if (this.enabled != wasEnabled) {
      this.cumulativeEffectiveAlphaChanged();
      this.requestRender();
    }
  }
  getEffectiveAlpha() {
    return this.alpha * (this.enabled ? 1.0 : kDimmedAlpha);
  }
  getCumulativeEffectiveAlpha() {
    if (this.cachedCumulativeEffectiveAlpha != undefined) {
      return this.cachedCumulativeEffectiveAlpha;
    }
    const myAlpha = this.getEffectiveAlpha();
    if (this.parentView) {
      return this.cachedCumulativeEffectiveAlpha = myAlpha * this.parentView.getCumulativeEffectiveAlpha();
    }
    return this.cachedCumulativeEffectiveAlpha = myAlpha;
  }
  cumulativeEffectiveAlphaChanged() {
    this.cachedCumulativeEffectiveAlpha = undefined;
    this.cachedOpaqueParent = undefined;
    for (const v of this.subViews.items) {
      v.cumulativeEffectiveAlphaChanged();
    }
  }
  getNearestOpaqueParent() {
    if (this.cachedOpaqueParent != undefined) {
      return this.cachedOpaqueParent;
    }
    let p = this.parentView;
    while (p != undefined) {
      if (p.getEffectiveAlpha() == 1.0) {
        return this.cachedOpaqueParent = p;
      }
      p = p.parentView;
    }
    return undefined;
  }
  findTop(r) {
    // find the top view that contains the rect
    if(this.containsR(r)) {
      var n = this.subViews.length();
      for(var i = 0; i < n; i++) {
        var v = this.subViews.get(i).findTop(r);
        if(v != undefined)
          return v;
      }

      return this;
    }
    return undefined;
  }
  offset(dx, dy) {
    super.offset(dx, dy);
    var n = this.subViews.length();
    for(var i = 0; i < n; i++)
      this.subViews.get(i).offset(dx, dy);
  }
  move(l, t) {
    this.offset(l - this.l, t - this.t);
  }
  setT(t) {
    this.offset(0, t - this.t);
  }
  setL(l) {
    this.offset(l - this.l, 0);
  }
  setLT(l, t) {
    this.offset(l - this.l, t - this.t);
  }

  requestRender() {
    if(this.root != undefined && this.isVis && this.isPVis) {
      if (this.getCumulativeEffectiveAlpha() != 1.0) {
        const p = this.getNearestOpaqueParent();
        if (p != undefined) {
          this.root.requestRender(p);
          return;
        }
      }
      this.root.requestRender(this);
    }
  }

  enableClipping(e) {
    this.isClippingEnabled = e;
  }

  setAlpha(a) {
    this.alpha = a;
    this.cumulativeEffectiveAlphaChanged();
  }

  renderSelfAndSubs(c) {
    if(this.isVis && this.isPVis) {
      c.save();

      if (c.globalAlpha != undefined) {
        c.globalAlpha *= this.getEffectiveAlpha();
      }

      if (this.isClippingEnabled) {
        // Clip the canvas to our rectangle
        let clipPath = new Path2D();
        clipPath.rect(this.l, this.t, this.w, this.h);
        c.clip(clipPath);
      }

      if(this.bg != undefined) {
        c.fillStyle = this.bg;
        c.fillRect(this.l, this.t, this.w, this.h);
      }

      this.render(c);

      // draw all the visible children
      var n = this.subViews.length();
      for(var i = 0; i < n; i++) {
        var v = this.subViews.get(i);
        v.renderSelfAndSubs(c);
      }

      c.restore();
      // this.dbgStrokeRect(c);
    }
  }

  setBG(bg) { this.bg = bg; }

  getMouseHandler(xy) {
    if(this.isVis && this.isPVis && this.enabled && this.contains(xy.x, xy.y)) {
      // find the top view that contains the coordinate
      var n = this.subViews.length();
      for(var i = 0; i < n; i++) {
        var v = this.subViews.get(i).getMouseHandler(xy);
        if(v != undefined)
          return v;
      }

      if(this.wantsMouse)
        return this;
    }
    return undefined;
  }

  dbgStrokeRect(c) {
    c.lineWidth = 1;
    c.strokeStyle = "yellow";
    c.strokeRect(this.l, this.t, this.w, this.h);
  }

  // extendend views implement render to draw their contents
  render(c) {}
  // extend these to handle the mouse
  handleDblClick() {}
  handleMouseDown(xy) {}
  handleMouseMoved(xy, delta) {}
  handleMouseUp() {}
}


