'use strict';

// Conversion between device gain and dB
function GainToDB(v) { return (v > 0) ? 20*Math.log10(v) : -Infinity; }
function DBToGain(dB) { return Math.pow(10, dB/20); }

// Warping between DB and the controls linear range of 0-1
const kCtrlRange = 72; // -Infinity
const kDBScale = 24.0;    // kDBScale and k0DBScalePos indicate where in the warping the specified DB lands. 

// Conversions for controls that scale 0 to 1 = -Infinity to 0dB
const k0DBScalePos = 0.33; // -24DB is 33% from the bottom.
const kExp = Math.log(1.0 - kDBScale / kCtrlRange) / Math.log(k0DBScalePos);
function CalcCtrl24To0dB(s) {
  return (s > 0) ? (kCtrlRange * (Math.pow(s, kExp) - 1.0)) : -Infinity;
}
function Calc0dBToCtrl24(dB) {
  if(dB > 0)
    return 1.0;
  
  if(dB < -(kCtrlRange - 1.0))
    return 0.0;

  return Math.pow((dB / kCtrlRange) + 1.0, 1.0 / kExp);
}


// Conversions for controls that scale 0 to 1 = -Infinity to +12dB
const k12DBScalePos = 0.66; // -24DB from the bottom.
const kCtrlRange12 = kCtrlRange+12;
const kExp12 = Math.log(1.0 - kDBScale / kCtrlRange12) / Math.log(k12DBScalePos);

function CalcCtrl24To12dB(s) {
  return (s > 0) ? ((kCtrlRange12 * (Math.pow(s, kExp12) - 1.0)) + 12) : -Infinity;
}
function Calc12dBToCtrl24(dB) {
  if(dB > 12.0)
    return 1.0;
  
  if(dB < -(kCtrlRange - 1.0))
    return 0.0;

  return Math.pow(((dB - 12) / kCtrlRange12) + 1.0, 1.0 / kExp12);
}

// Linear conversion for controls that scale 0 to 1 = -72 to 0 dB
function CalcDBToCtrlLinear(dB) {
  // for Dynamics meters the scale is linear -72 to 0
  return 1 - ((dB < -71) ? 1 : -dB/71);
}

// For drawing tick marks
function drawTicks(c, color, ticks, t, l, h, r) {
  t += 2;
  c.lineWidth = 1;
  c.lineCap = "butt";
  c.fillStyle = c.strokeStyle = color;
  c.font = "10px Helvetica";
  c.textAlign = 'right';

  c.beginPath();
  var numT = ticks.length;
  for (var i = 0; i < numT; i++) {
    var y = t + (h * ticks[i].s);
    c.moveTo(l+21, y);
    c.lineTo(r, y);
    c.fillText(ticks[i].l, l+19, y + 3);
  }
  c.stroke();
}

// For drawing tick marks
function drawTicksNoText(c, color, ticks, t, l, h) {
  t += 2;
  c.lineWidth = 1;
  c.lineCap = "butt";
  c.fillStyle = c.strokeStyle = color;

  c.beginPath();
  var numT = ticks.length;
  for (var i = 0; i < numT; i++) {
    var y = t + (h * ticks[i].s);
    c.moveTo(l, y);
    c.lineTo(l+16, y);
  }
  c.stroke();
}

const k0dBTicks = [
  { l:'0', s:1-Calc0dBToCtrl24(0) },
  { l:'-6', s:1-Calc0dBToCtrl24(-6) },
  { l:'-12', s:1-Calc0dBToCtrl24(-12) },
  { l:'-24', s:1-Calc0dBToCtrl24(-24) },
  { l:'-48', s:1-Calc0dBToCtrl24(-48) },
  { l:'-∞', s:1-Calc0dBToCtrl24(-kCtrlRange) },
];
const k12dBTicks = [
  { l:'+12', s:1-Calc12dBToCtrl24(12) },
  { l:'+6', s:1-Calc12dBToCtrl24(6) },
  { l:'0', s:1-Calc12dBToCtrl24(0) },
  { l:'-6', s:1-Calc12dBToCtrl24(-6) },
  { l:'-12', s:1-Calc12dBToCtrl24(-12) },
  { l:'-18', s:1-Calc12dBToCtrl24(-18) },
  { l:'-24', s:1-Calc12dBToCtrl24(-24) },
  { l:'-32', s:1-Calc12dBToCtrl24(-32) },
  { l:'-48', s:1-Calc12dBToCtrl24(-48) },
  { l:'-∞', s:1-Calc12dBToCtrl24(-kCtrlRange) },
];
function draw0dBTicks(c, color, t, l, h) {
  drawTicks(c, color, k0dBTicks, t, l, h);
}
function draw12dBTicks(c, color, t, l, h, r) {
  drawTicks(c, color, k12dBTicks, t, l, h, r);
}
function draw12dBTicksNoText(c, color, t, l, h) {
  drawTicksNoText(c, color, k12dBTicks, t, l, h);
}
